import pandas as pd
import re
from .constants import (ADVANCED_OPPONENT_STATS_URL,
                        ADVANCED_STATS_URL,
                        BASIC_OPPONENT_STATS_URL,
                        BASIC_STATS_URL,
                        PARSING_SCHEME)
from pyquery import PyQuery as pq
from ..decorators import float_property_decorator, int_property_decorator
from .. import utils
from .conferences import Conferences
from .roster import Roster
from .schedule import Schedule


class Team(object):
    """
    An object containing all of a team's season information.

    Finds and parses all team stat information and identifiers, such as full
    and short names, and sets them as properties which can be directly read
    from for easy reference.

    Parameters
    ----------
    team_data : string
        A string containing all of the rows of stats for a given team. If
        multiple tables are being referenced, this will be comprised of
        multiple rows in a single string.
    team_conference : string (optional)
        A string of the team's conference abbreviation, such as 'big-12'.
    year : string (optional)
        The requested year to pull stats from.
    """
    def __init__(self, team_data, team_conference=None, year=None):
        self._team_conference = team_conference
        self._year = year
        self._abbreviation = None
        self._name = None
        self._games_played = None
        self._wins = None
        self._losses = None
        self._win_percentage = None
        self._simple_rating_system = None
        self._strength_of_schedule = None
        self._conference_wins = None
        self._conference_losses = None
        self._home_wins = None
        self._home_losses = None
        self._away_wins = None
        self._away_losses = None
        self._points = None
        self._opp_points = None
        self._minutes_played = None
        self._field_goals = None
        self._field_goal_attempts = None
        self._field_goal_percentage = None
        self._two_point_field_goals = None
        self._two_point_field_goal_attempts = None
        self._two_point_field_goal_percentage = None
        self._three_point_field_goals = None
        self._three_point_field_goal_attempts = None
        self._three_point_field_goal_percentage = None
        self._free_throws = None
        self._free_throw_attempts = None
        self._free_throw_percentage = None
        self._offensive_rebounds = None
        self._defensive_rebounds = None
        self._total_rebounds = None
        self._assists = None
        self._steals = None
        self._blocks = None
        self._turnovers = None
        self._personal_fouls = None
        self._opp_field_goals = None
        self._opp_field_goal_attempts = None
        self._opp_field_goal_percentage = None
        self._opp_two_point_field_goals = None
        self._opp_two_point_field_goal_attempts = None
        self._opp_two_point_field_goal_percentage = None
        self._opp_three_point_field_goals = None
        self._opp_three_point_field_goal_attempts = None
        self._opp_three_point_field_goal_percentage = None
        self._opp_free_throws = None
        self._opp_free_throw_attempts = None
        self._opp_free_throw_percentage = None
        self._opp_offensive_rebounds = None
        self._opp_defensive_rebounds = None
        self._opp_total_rebounds = None
        self._opp_assists = None
        self._opp_steals = None
        self._opp_blocks = None
        self._opp_turnovers = None
        self._opp_personal_fouls = None
        self._pace = None
        self._offensive_rating = None
        self._free_throw_attempt_rate = None
        self._three_point_attempt_rate = None
        self._true_shooting_percentage = None
        self._total_rebound_percentage = None
        self._assist_percentage = None
        self._steal_percentage = None
        self._block_percentage = None
        self._effective_field_goal_percentage = None
        self._turnover_percentage = None
        self._offensive_rebound_percentage = None
        self._free_throws_per_field_goal_attempt = None
        self._opp_offensive_rating = None
        self._opp_free_throw_attempt_rate = None
        self._opp_three_point_attempt_rate = None
        self._opp_true_shooting_percentage = None
        self._opp_total_rebound_percentage = None
        self._opp_assist_percentage = None
        self._opp_steal_percentage = None
        self._opp_block_percentage = None
        self._opp_effective_field_goal_percentage = None
        self._opp_turnover_percentage = None
        self._opp_offensive_rebound_percentage = None
        self._opp_free_throws_per_field_goal_attempt = None

        self._parse_team_data(team_data)

    def _parse_team_data(self, team_data):
        """
        Parses a value for every attribute.

        This function looks through every attribute and retrieves the value
        according to the parsing scheme and index of the attribute from the
        passed HTML data. Once the value is retrieved, the attribute's value is
        updated with the returned result.

        Note that this method is called directly once Team is invoked and does
        not need to be called manually.

        Parameters
        ----------
        team_data : string
            A string containing all of the rows of stats for a given team. If
            multiple tables are being referenced, this will be comprised of
            multiple rows in a single string.
        """
        for field in self.__dict__:
            if field == '_year' or \
               field == '_team_conference':
                continue
            value = utils._parse_field(PARSING_SCHEME,
                                       team_data,
                                       # Remove the '_' from the name
                                       str(field)[1:])
            setattr(self, field, value)

    @property
    def dataframe(self):
        """
        Returns a pandas DataFrame containing all other class properties and
        values. The index for the DataFrame is the string abbreviation of the
        team, such as 'PURDUE'.
        """
        fields_to_include = {
            'abbreviation': self.abbreviation,
            'assist_percentage': self.assist_percentage,
            'assists': self.assists,
            'away_losses': self.away_losses,
            'away_wins': self.away_wins,
            'block_percentage': self.block_percentage,
            'blocks': self.blocks,
            'conference': self.conference,
            'conference_losses': self.conference_losses,
            'conference_wins': self.conference_wins,
            'defensive_rebounds': self.defensive_rebounds,
            'effective_field_goal_percentage':
            self.effective_field_goal_percentage,
            'field_goal_attempts': self.field_goal_attempts,
            'field_goal_percentage': self.field_goal_percentage,
            'field_goals': self.field_goals,
            'free_throw_attempt_rate': self.free_throw_attempt_rate,
            'free_throw_attempts': self.free_throw_attempts,
            'free_throw_percentage': self.free_throw_percentage,
            'free_throws': self.free_throws,
            'free_throws_per_field_goal_attempt':
            self.free_throws_per_field_goal_attempt,
            'games_played': self.games_played,
            'home_losses': self.home_losses,
            'home_wins': self.home_wins,
            'losses': self.losses,
            'minutes_played': self.minutes_played,
            'name': self.name,
            'net_rating': self.net_rating,
            'offensive_rating': self.offensive_rating,
            'offensive_rebound_percentage': self.offensive_rebound_percentage,
            'offensive_rebounds': self.offensive_rebounds,
            'opp_assist_percentage': self.opp_assist_percentage,
            'opp_assists': self.opp_assists,
            'opp_block_percentage': self.opp_block_percentage,
            'opp_blocks': self.opp_blocks,
            'opp_defensive_rebounds': self.opp_defensive_rebounds,
            'opp_effective_field_goal_percentage':
            self.opp_effective_field_goal_percentage,
            'opp_field_goal_attempts': self.opp_field_goal_attempts,
            'opp_field_goal_percentage': self.opp_field_goal_percentage,
            'opp_field_goals': self.opp_field_goals,
            'opp_free_throw_attempt_rate': self.opp_free_throw_attempt_rate,
            'opp_free_throw_attempts': self.opp_free_throw_attempts,
            'opp_free_throw_percentage': self.opp_free_throw_percentage,
            'opp_free_throws': self.opp_free_throws,
            'opp_free_throws_per_field_goal_attempt':
            self.opp_free_throws_per_field_goal_attempt,
            'opp_offensive_rating': self.opp_offensive_rating,
            'opp_offensive_rebound_percentage':
            self.opp_offensive_rebound_percentage,
            'opp_offensive_rebounds': self.opp_offensive_rebounds,
            'opp_personal_fouls': self.opp_personal_fouls,
            'opp_points': self.opp_points,
            'opp_steal_percentage': self.opp_steal_percentage,
            'opp_steals': self.opp_steals,
            'opp_three_point_attempt_rate': self.opp_three_point_attempt_rate,
            'opp_three_point_field_goal_attempts':
            self.opp_three_point_field_goal_attempts,
            'opp_three_point_field_goal_percentage':
            self.opp_three_point_field_goal_percentage,
            'opp_three_point_field_goals': self.opp_three_point_field_goals,
            'opp_two_point_field_goal_attempts':
            self.opp_two_point_field_goal_attempts,
            'opp_two_point_field_goal_percentage':
            self.opp_two_point_field_goal_percentage,
            'opp_two_point_field_goals': self.opp_two_point_field_goals,
            'opp_total_rebound_percentage': self.opp_total_rebound_percentage,
            'opp_total_rebounds': self.opp_total_rebounds,
            'opp_true_shooting_percentage': self.opp_true_shooting_percentage,
            'opp_turnover_percentage': self.opp_turnover_percentage,
            'opp_turnovers': self.opp_turnovers,
            'pace': self.pace,
            'personal_fouls': self.personal_fouls,
            'points': self.points,
            'simple_rating_system': self.simple_rating_system,
            'steal_percentage': self.steal_percentage,
            'steals': self.steals,
            'strength_of_schedule': self.strength_of_schedule,
            'three_point_attempt_rate': self.three_point_attempt_rate,
            'three_point_field_goal_attempts':
            self.three_point_field_goal_attempts,
            'three_point_field_goal_percentage':
            self.three_point_field_goal_percentage,
            'three_point_field_goals': self.three_point_field_goals,
            'two_point_field_goal_attempts':
            self.two_point_field_goal_attempts,
            'two_point_field_goal_percentage':
            self.two_point_field_goal_percentage,
            'two_point_field_goals': self.two_point_field_goals,
            'total_rebound_percentage': self.total_rebound_percentage,
            'total_rebounds': self.total_rebounds,
            'true_shooting_percentage': self.true_shooting_percentage,
            'turnover_percentage': self.turnover_percentage,
            'turnovers': self.turnovers,
            'win_percentage': self.win_percentage,
            'wins': self.wins
        }
        return pd.DataFrame([fields_to_include], index=[self._abbreviation])

    @property
    def conference(self):
        """
        Returns a ``string`` of the team's conference abbreviation, such as
        'big-12' for the Big 12 Conference.
        """
        return self._team_conference

    @property
    def abbreviation(self):
        """
        Returns a ``string`` of the team's short name, such as 'PURDUE' for the
        Purdue Boilermakers.
        """
        return self._abbreviation

    @property
    def schedule(self):
        """
        Returns an instance of the Schedule class containing the team's
        complete schedule for the season.
        """
        return Schedule(self._abbreviation, self._year)

    @property
    def roster(self):
        """
        Returns an instance of the Roster class containing all players for the
        team during the season with all career stats.
        """
        return Roster(self._abbreviation, self._year)

    @property
    def name(self):
        """
        Returns a ``string`` of the team's full name, such as 'Purdue
        Boilermakers'.
        """
        return self._name

    @int_property_decorator
    def games_played(self):
        """
        Returns an ``int`` of the total number of games the team has played
        during the season.
        """
        return self._games_played

    @int_property_decorator
    def wins(self):
        """
        Returns an ``int`` of the total number of games the team won during the
        season.
        """
        return self._wins

    @int_property_decorator
    def losses(self):
        """
        Returns an ``int`` of the total number of games the team lost during
        the season.
        """
        return self._losses

    @float_property_decorator
    def win_percentage(self):
        """
        Returns a ``float`` of the number of wins divided by the number of
        games played during the season. Percentage ranges from 0-1.
        """
        return self._win_percentage

    @float_property_decorator
    def simple_rating_system(self):
        """
        Returns a ``float`` of the team's average point differential compared
        to the strength of schedule. Higher values indicate stronger teams. An
        average team is denoted with 0.0. Negative numbers are comparatively
        worse than average.
        """
        return self._simple_rating_system

    @float_property_decorator
    def strength_of_schedule(self):
        """
        Returns a ``float`` of the team's strength of schedule based on the
        points above and below average. An average strength of schedule is
        denoted with 0.0. Negative numbers are comparatively easier than
        average.
        """
        return self._strength_of_schedule

    @int_property_decorator
    def conference_wins(self):
        """
        Returns an ``int`` of the total number of conference games the team won
        during the season.
        """
        return self._conference_wins

    @int_property_decorator
    def conference_losses(self):
        """
        Returns an ``int`` of the total number of conference games the team
        lost during the season.
        """
        return self._conference_losses

    @int_property_decorator
    def home_wins(self):
        """
        Returns an ``int`` of the total number of home games the team won
        during the season.
        """
        return self._home_wins

    @int_property_decorator
    def home_losses(self):
        """
        Returns an ``int`` of the total number of home games the team lost
        during the season.
        """
        return self._home_losses

    @int_property_decorator
    def away_wins(self):
        """
        Returns an ``int`` of the total number of away games the team won
        during the season.
        """
        return self._away_wins

    @int_property_decorator
    def away_losses(self):
        """
        Returns an ``int`` of the total number of away games the team lost
        during the season.
        """
        return self._away_losses

    @int_property_decorator
    def points(self):
        """
        Returns an ``int`` of the total number of points the team scored during
        the season.
        """
        return self._points

    @int_property_decorator
    def opp_points(self):
        """
        Returns an ``int`` of the total number of points opponents have scored
        during the season.
        """
        return self._opp_points

    @int_property_decorator
    def minutes_played(self):
        """
        Returns an ``int`` of the total number of minutes played by the team
        during the season.
        """
        return self._minutes_played

    @int_property_decorator
    def field_goals(self):
        """
        Returns an ``int`` of the total number of field goals made during the
        season.
        """
        return self._field_goals

    @int_property_decorator
    def field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of field goal attempts during
        the season.
        """
        return self._field_goal_attempts

    @float_property_decorator
    def field_goal_percentage(self):
        """
        Returns a ``float`` of the number of field goals made divided by the
        total number of field goal attempts. Percentage ranges from 0-1.
        """
        return self._field_goal_percentage

    @int_property_decorator
    def two_point_field_goals(self):
        """
        Returns an ``int`` of the total number of two point field goals made
        during the season.
        """
        return self.field_goals - self.three_point_field_goals

    @int_property_decorator
    def two_point_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of two point field goal attempts
        during the season.
        """
        return self.field_goal_attempts - self.three_point_field_goal_attempts

    @float_property_decorator
    def two_point_field_goal_percentage(self):
        """
        Returns a ``float`` of the number of two point field goals made divided
        by the number of two point field goal attempts. Percentage ranges from
        0-1.
        """
        try:
            result = float(self.two_point_field_goals) / \
                float(self.two_point_field_goal_attempts)
            return round(result, 3)
        except ZeroDivisionError:
            return 0.0

    @int_property_decorator
    def three_point_field_goals(self):
        """
        Returns an ``int`` of the total number of three point field goals made
        during the season.
        """
        return self._three_point_field_goals

    @int_property_decorator
    def three_point_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of three point field goal
        attempts during the season.
        """
        return self._three_point_field_goal_attempts

    @float_property_decorator
    def three_point_field_goal_percentage(self):
        """
        Returns a ``float`` of the number of three point field goals made
        divided by the number of three point field goal attempts. Percentage
        ranges from 0-1.
        """
        return self._three_point_field_goal_percentage

    @int_property_decorator
    def free_throws(self):
        """
        Returns an ``int`` of the total number of free throws made during the
        season.
        """
        return self._free_throws

    @int_property_decorator
    def free_throw_attempts(self):
        """
        Returns an ``int`` of the total number of free throw attempts during
        the season.
        """
        return self._free_throw_attempts

    @float_property_decorator
    def free_throw_percentage(self):
        """
        Returns a ``float`` of the number of free throws made divided by the
        number of free throw attempts during the season.
        """
        return self._free_throw_percentage

    @int_property_decorator
    def offensive_rebounds(self):
        """
        Returns an ``int`` of the total number of offensive rebounds during the
        season.
        """
        return self._offensive_rebounds

    @int_property_decorator
    def defensive_rebounds(self):
        """
        Returns an ``int`` of the total number of defensive rebounds during the
        season.
        """
        return self.total_rebounds - self.offensive_rebounds

    @int_property_decorator
    def total_rebounds(self):
        """
        Returns an ``int`` of the total number of rebounds during the season.
        """
        return self._total_rebounds

    @int_property_decorator
    def assists(self):
        """
        Returns an ``int`` of the total number of assists during the season.
        """
        return self._assists

    @int_property_decorator
    def steals(self):
        """
        Returns an ``int`` of the total number of steals during the season.
        """
        return self._steals

    @int_property_decorator
    def blocks(self):
        """
        Returns an ``int`` of the total number of blocks during the season.
        """
        return self._blocks

    @int_property_decorator
    def turnovers(self):
        """
        Returns an ``int`` of the total number of turnovers during the season.
        """
        return self._turnovers

    @int_property_decorator
    def personal_fouls(self):
        """
        Returns an ``int`` of the total number of personal fouls during the
        season.
        """
        return self._personal_fouls

    @int_property_decorator
    def opp_field_goals(self):
        """
        Returns an ``int`` of the total number of field goals made during the
        season by opponents.
        """
        return self._opp_field_goals

    @int_property_decorator
    def opp_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of field goal attempts during
        the season by opponents.
        """
        return self._opp_field_goal_attempts

    @float_property_decorator
    def opp_field_goal_percentage(self):
        """
        Returns a ``float`` of the number of field goals made divided by the
        total number of field goal attempts by opponents. Percentage ranges
        from 0-1.
        """
        return self._opp_field_goal_percentage

    @int_property_decorator
    def opp_two_point_field_goals(self):
        """
        Returns an ``int`` of the total number of two point field goals made
        during the season by opponents.
        """
        return self.opp_field_goals - self.opp_three_point_field_goals

    @int_property_decorator
    def opp_two_point_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of two point field goal attempts
        during the season by opponents.
        """
        return self.opp_field_goal_attempts - \
            self.opp_three_point_field_goal_attempts

    @float_property_decorator
    def opp_two_point_field_goal_percentage(self):
        """
        Returns a ``float`` of the number of two point field goals made divided
        by the number of two point field goal attempts by opponents. Percentage
        ranges from 0-1.
        """
        try:
            result = float(self.opp_two_point_field_goals) / \
                float(self.opp_two_point_field_goal_attempts)
            return round(result, 3)
        except ZeroDivisionError:
            return 0.0

    @int_property_decorator
    def opp_three_point_field_goals(self):
        """
        Returns an ``int`` of the total number of three point field goals made
        during the season by opponents.
        """
        return self._opp_three_point_field_goals

    @int_property_decorator
    def opp_three_point_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of three point field goal
        attempts during the season by opponents.
        """
        return self._opp_three_point_field_goal_attempts

    @float_property_decorator
    def opp_three_point_field_goal_percentage(self):
        """
        Returns a ``float`` of the number of three point field goals made
        divided by the number of three point field goal attempts by opponents.
        Percentage ranges from 0-1.
        """
        return self._opp_three_point_field_goal_percentage

    @int_property_decorator
    def opp_free_throws(self):
        """
        Returns an ``int`` of the total number of free throws made during the
        season by opponents.
        """
        return self._opp_free_throws

    @int_property_decorator
    def opp_free_throw_attempts(self):
        """
        Returns an ``int`` of the total number of free throw attempts during
        the season by opponents.
        """
        return self._opp_free_throw_attempts

    @float_property_decorator
    def opp_free_throw_percentage(self):
        """
        Returns a ``float`` of the number of free throws made divided by the
        number of free throw attempts during the season by opponents.
        """
        return self._opp_free_throw_percentage

    @int_property_decorator
    def opp_offensive_rebounds(self):
        """
        Returns an ``int`` of the total number of offensive rebounds during the
        season by opponents.
        """
        return self._opp_offensive_rebounds

    @int_property_decorator
    def opp_defensive_rebounds(self):
        """
        Returns an ``int`` of the total number of defensive rebounds during the
        season by opponents.
        """
        return self.opp_total_rebounds - self.opp_offensive_rebounds

    @int_property_decorator
    def opp_total_rebounds(self):
        """
        Returns an ``int`` of the total number of rebounds during the season by
        opponents.
        """
        return self._opp_total_rebounds

    @int_property_decorator
    def opp_assists(self):
        """
        Returns an ``int`` of the total number of assists during the season by
        opponents.
        """
        return self._opp_assists

    @int_property_decorator
    def opp_steals(self):
        """
        Returns an ``int`` of the total number of steals during the season by
        opponents.
        """
        return self._opp_steals

    @int_property_decorator
    def opp_blocks(self):
        """
        Returns an ``int`` of the total number of blocks during the season by
        opponents.
        """
        return self._opp_blocks

    @int_property_decorator
    def opp_turnovers(self):
        """
        Returns an ``int`` of the total number of turnovers during the season
        by opponents.
        """
        return self._opp_turnovers

    @int_property_decorator
    def opp_personal_fouls(self):
        """
        Returns an ``int`` of the total number of personal fouls during the
        season by opponents.
        """
        return self._opp_personal_fouls

    @float_property_decorator
    def pace(self):
        """
        Returns a ``float`` of the average number of possessions per 40
        minutes.
        """
        return self._pace

    @float_property_decorator
    def offensive_rating(self):
        """
        Returns a ``float`` of the average number of points scored per 100
        possessions.
        """
        return self._offensive_rating

    @float_property_decorator
    def net_rating(self):
        """
        Returns a ``float`` of the net team rating which is equivalent to the
        difference between the offensive rating and the defensive (or the
        opponent's offensive) rating. Positive values indicate teams that score
        more points than they allow per 100 possessions.
        """
        return self.offensive_rating - self.opp_offensive_rating

    @float_property_decorator
    def free_throw_attempt_rate(self):
        """
        Returns a ``float`` of the average number of free throw attempts per
        field goal attempt.
        """
        return self._free_throw_attempt_rate

    @float_property_decorator
    def three_point_attempt_rate(self):
        """
        Returns a ``float`` of the percentage of field goal attempts from
        3-point range. Percentage ranges from 0-1.
        """
        return self._three_point_attempt_rate

    @float_property_decorator
    def true_shooting_percentage(self):
        """
        Returns a ``float`` of the team's true shooting percentage which
        considers free throws, 2-point field goals, and 3-point field goals.
        Percentage ranges from 0-1.
        """
        return self._true_shooting_percentage

    @float_property_decorator
    def total_rebound_percentage(self):
        """
        Returns a ``float`` of the percentage of available rebounds a team
        grabbed.
        Percentage ranges from 0-100.
        """
        return self._total_rebound_percentage

    @float_property_decorator
    def assist_percentage(self):
        """
        Returns a ``float`` of the percentage of field goals that were
        assisted.
        Percentage ranges from 0-100.
        """
        return self._assist_percentage

    @float_property_decorator
    def steal_percentage(self):
        """
        Returns a ``float`` of the percentage of opponent possessions that
        ended in a steal. Percentage ranges from 0-100.
        """
        return self._steal_percentage

    @float_property_decorator
    def block_percentage(self):
        """
        Returns a ``float`` of the percentage of 2-point field goals by the
        opponent that were blocked. Percentage ranges from 0-100.
        """
        return self._block_percentage

    @float_property_decorator
    def effective_field_goal_percentage(self):
        """
        Returns a ``float`` of the field goal percentage while giving extra
        weight to 3-point field goals. Percentage ranges from 0-1.
        """
        return self._effective_field_goal_percentage

    @float_property_decorator
    def turnover_percentage(self):
        """
        Returns a ``float`` of the number of times the team turned the ball
        over per 100 possessions.
        """
        return self._turnover_percentage

    @float_property_decorator
    def offensive_rebound_percentage(self):
        """
        Returns a ``float`` of the percentage of available offensive rebounds a
        team grabbed. Percentage ranges from 0-100.
        """
        return self._offensive_rebound_percentage

    @float_property_decorator
    def free_throws_per_field_goal_attempt(self):
        """
        Returns a ``float`` of the number of free throws per field goal
        attempt.
        """
        return self._free_throws_per_field_goal_attempt

    @float_property_decorator
    def opp_offensive_rating(self):
        """
        Returns a ``float`` of the average number of points scored per 100
        possessions by the opponent. This is equivalent to the team's defensive
        rating as it is the number of points the team allows per 100
        possessions by the opponent.
        """
        return self._opp_offensive_rating

    @float_property_decorator
    def opp_free_throw_attempt_rate(self):
        """
        Returns a ``float`` of the average number of free throw attempts per
        field goal attempt by the opponent.
        """
        return self._opp_free_throw_attempt_rate

    @float_property_decorator
    def opp_three_point_attempt_rate(self):
        """
        Returns a ``float`` of the percentage of field goal attempts from
        3-point range by the opponent. Percentage ranges from 0-1.
        """
        return self._opp_three_point_attempt_rate

    @float_property_decorator
    def opp_true_shooting_percentage(self):
        """
        Returns a ``float`` of the opponent's true shooting percentage which
        considers free throws, 2-point field goals, and 3-point field goals.
        Percentage ranges from 0-1.
        """
        return self._opp_true_shooting_percentage

    @float_property_decorator
    def opp_total_rebound_percentage(self):
        """
        Returns a ``float`` of the percentage of available rebounds the
        opponent grabbed. Percentage ranges from 0-100.
        """
        return self._opp_total_rebound_percentage

    @float_property_decorator
    def opp_assist_percentage(self):
        """
        Returns a ``float`` of the percentage of the opponent's field goals
        that were assisted. Percentage ranges from 0-100.
        """
        return self._opp_assist_percentage

    @float_property_decorator
    def opp_steal_percentage(self):
        """
        Returns a ``float`` of the percentage of possessions that ended in a
        steal by the opponent. Percentage ranges from 0-100.
        """
        return self._opp_steal_percentage

    @float_property_decorator
    def opp_block_percentage(self):
        """
        Returns a ``float`` of the percentage of 2-point field goals that were
        blocked by the opponent. Percentage ranges from 0-100.
        """
        return self._opp_block_percentage

    @float_property_decorator
    def opp_effective_field_goal_percentage(self):
        """
        Returns a ``float`` of the opponent's field goal percentage while
        giving extra weight to 3-point field goals. Percentage ranges from 0-1.
        """
        return self._opp_effective_field_goal_percentage

    @float_property_decorator
    def opp_turnover_percentage(self):
        """
        Returns a ``float`` of the number of times the opponent turned the ball
        over per 100 possessions.
        """
        return self._opp_turnover_percentage

    @float_property_decorator
    def opp_offensive_rebound_percentage(self):
        """
        Returns a ``float`` of the percentage of available offensive rebounds
        the opponent grabbed. Percentage ranges from 0-100.
        """
        return self._opp_offensive_rebound_percentage

    @float_property_decorator
    def opp_free_throws_per_field_goal_attempt(self):
        """
        Returns a ``float`` of the number of free throws per field goal attempt
        by the opponent.
        """
        return self._opp_free_throws_per_field_goal_attempt


class Teams:
    """
    A list of all NCAA Men's Basketball teams and their stats in a given year.

    Finds and retrieves a list of all NCAA Men's Basketball teams from
    www.sports-reference.com and creates a Team instance for every team that
    participated in the league in a given year. The Team class comprises a list
    of all major stats and a few identifiers for the requested season.

    Parameters
    ----------
    year : string (optional)
        The requested year to pull stats from.
    """
    def __init__(self, year=None):
        self._teams = []
        self._conferences_dict = Conferences(year).team_conference

        self._retrieve_all_teams(year)

    def __getitem__(self, abbreviation):
        """
        Return a specified team.

        Returns a team's instance in the Teams class as specified by the team's
        short name.

        Parameters
        ----------
        abbreviation : string
            An NCAAB team's short name (ie. 'PURDUE' for the Purdue
            Boilermakers).

        Returns
        -------
        Team instance
            If the requested team can be found, its Team instance is returned.

        Raises
        ------
        ValueError
            If the requested team is not present within the Teams list.
        """
        for team in self._teams:
            if team.abbreviation.upper() == abbreviation.upper():
                return team
        raise ValueError('Team abbreviation %s not found' % abbreviation)

    def __call__(self, abbreviation):
        """
        Return a specified team.

        Returns a team's instance in the Teams class as specified by the team's
        short name. This method is a wrapper for __getitem__.

        Parameters
        ----------
        abbreviation : string
            An NCAAB team's short name (ie. 'PURDUE' for the Purdue
            Boilermakers).

        Returns
        -------
        Team instance
            if the requested team can be found, its Team instance is returned
        """
        return self.__getitem__(abbreviation)

    def __repr__(self):
        """Returns a ``list`` of all NCAAB teams for the given season."""
        return self._teams

    def __iter__(self):
        """Returns an iterator of all of the NCAAB teams for a given season."""
        return iter(self.__repr__())

    def __len__(self):
        """Returns the number of NCAAB teams for a given season."""
        return len(self.__repr__())

    def _add_stats_data(self, teams_list, team_data_dict):
        """
        Add a team's stats row to a dictionary.

        Pass table contents and a stats dictionary of all teams to accumulate
        all stats for each team in a single variable.

        Parameters
        ----------
        teams_list : generator
            A generator of all row items in a given table.
        team_data_dict : {str: {'data': str}} dictionary
            A dictionary where every key is the team's abbreviation and every
            value is another dictionary with a 'data' key which contains the
            string version of the row data for the matched team.

        Returns
        -------
        dictionary
            An updated version of the team_data_dict with the passed table row
            information included.
        """
        for team_data in teams_list:
            if 'class="over_header thead"' in str(team_data) or\
               'class="thead"' in str(team_data):
                continue
            abbr = utils._parse_field(PARSING_SCHEME,
                                      team_data,
                                      'abbreviation')
            try:
                team_data_dict[abbr]['data'] += team_data
            except KeyError:
                team_data_dict[abbr] = {'data': team_data}
        return team_data_dict

    def _retrieve_all_teams(self, year):
        """
        Find and create Team instances for all teams in the given season.

        For a given season, parses the specified NCAAB stats table and finds
        all requested stats. Each team then has a Team instance created which
        includes all requested stats and a few identifiers, such as the team's
        name and abbreviation. All of the individual Team instances are added
        to a list.

        Note that this method is called directly once Teams is invoked and does
        not need to be called manually.

        Parameters
        ----------
        year : string
            The requested year to pull stats from.
        """
        team_data_dict = {}

        if not year:
            year = utils._find_year_for_season('ncaab')
        doc = pq(BASIC_STATS_URL % year)
        teams_list = utils._get_stats_table(doc, 'table#basic_school_stats')
        doc = pq(BASIC_OPPONENT_STATS_URL % year)
        opp_list = utils._get_stats_table(doc, 'table#basic_opp_stats')
        doc = pq(ADVANCED_STATS_URL % year)
        adv_teams_list = utils._get_stats_table(doc, 'table#adv_school_stats')
        doc = pq(ADVANCED_OPPONENT_STATS_URL % year)
        adv_opp_list = utils._get_stats_table(doc, 'table#adv_opp_stats')

        for stats_list in [teams_list, opp_list, adv_teams_list, adv_opp_list]:
            team_data_dict = self._add_stats_data(stats_list, team_data_dict)

        for team_name, team_data in team_data_dict.items():
            team = Team(team_data['data'],
                        self._conferences_dict[team_name.lower()],
                        year)
            self._teams.append(team)

    @property
    def dataframes(self):
        """
        Returns a pandas DataFrame where each row is a representation of the
        Team class. Rows are indexed by the team abbreviation.
        """
        frames = []
        for team in self.__iter__():
            frames.append(team.dataframe)
        return pd.concat(frames)



PARSING_SCHEME = {
    'name': 'a',
    'games_played': 'td[data-stat="g"]:first',
    'minutes_played': 'td[data-stat="mp"]:first',
    'field_goals': 'td[data-stat="fg"]:first',
    'field_goal_attempts': 'td[data-stat="fga"]:first',
    'field_goal_percentage': 'td[data-stat="fg_pct"]:first',
    'three_point_field_goals': 'td[data-stat="fg3"]:first',
    'three_point_field_goal_attempts': 'td[data-stat="fg3a"]:first',
    'three_point_field_goal_percentage': 'td[data-stat="fg3_pct"]:first',
    'two_point_field_goals': 'td[data-stat="fg2"]:first',
    'two_point_field_goal_attempts': 'td[data-stat="fg2a"]:first',
    'two_point_field_goal_percentage': 'td[data-stat="fg2_pct"]:first',
    'free_throws': 'td[data-stat="ft"]:first',
    'free_throw_attempts': 'td[data-stat="fta"]:first',
    'free_throw_percentage': 'td[data-stat="ft_pct"]:first',
    'offensive_rebounds': 'td[data-stat="orb"]:first',
    'defensive_rebounds': 'td[data-stat="drb"]:first',
    'total_rebounds': 'td[data-stat="trb"]:first',
    'assists': 'td[data-stat="ast"]:first',
    'steals': 'td[data-stat="stl"]:first',
    'blocks': 'td[data-stat="blk"]:first',
    'turnovers': 'td[data-stat="tov"]:first',
    'personal_fouls': 'td[data-stat="pf"]:first',
    'points': 'td[data-stat="pts"]:first',
    'opp_minutes_played': 'td[data-stat="mp"]:first',
    'opp_field_goals': 'td[data-stat="opp_fg"]:first',
    'opp_field_goal_attempts': 'td[data-stat="opp_fga"]:first',
    'opp_field_goal_percentage': 'td[data-stat="opp_fg_pct"]:first',
    'opp_three_point_field_goals': 'td[data-stat="opp_fg3"]:first',
    'opp_three_point_field_goal_attempts': 'td[data-stat="opp_fg3a"]:first',
    'opp_three_point_field_goal_percentage':
    'td[data-stat="opp_fg3_pct"]:first',
    'opp_two_point_field_goals': 'td[data-stat="opp_fg2"]:first',
    'opp_two_point_field_goal_attempts': 'td[data-stat="opp_fg2a"]:first',
    'opp_two_point_field_goal_percentage': 'td[data-stat="opp_fg2_pct"]:first',
    'opp_free_throws': 'td[data-stat="opp_ft"]:first',
    'opp_free_throw_attempts': 'td[data-stat="opp_fta"]:first',
    'opp_free_throw_percentage': 'td[data-stat="opp_ft_pct"]:first',
    'opp_offensive_rebounds': 'td[data-stat="opp_orb"]:first',
    'opp_defensive_rebounds': 'td[data-stat="opp_drb"]:first',
    'opp_total_rebounds': 'td[data-stat="opp_trb"]:first',
    'opp_assists': 'td[data-stat="opp_ast"]:first',
    'opp_steals': 'td[data-stat="opp_stl"]:first',
    'opp_blocks': 'td[data-stat="opp_blk"]:first',
    'opp_turnovers': 'td[data-stat="opp_tov"]:first',
    'opp_personal_fouls': 'td[data-stat="opp_pf"]:first',
    'opp_points': 'td[data-stat="opp_pts"]:first'
}

SCHEDULE_SCHEME = {
    'game': 'th[data-stat="g"]:first',
    'date': 'td[data-stat="date_game"]:first',
    'time': 'td[data-stat="game_start_time"]:first',
    'boxscore': 'td[data-stat="box_score_text"]:first',
    'location': 'td[data-stat="game_location"]:first',
    'opponent_abbr': 'td[data-stat="opp_id"]:first',
    'opponent_name': 'td[data-stat="opp_name"]:first',
    'result': 'td[data-stat="game_result"]:first',
    'points_scored': 'td[data-stat="pts"]:first',
    'points_allowed': 'td[data-stat="opp_pts"]:first',
    'wins': 'td[data-stat="wins"]:first',
    'losses': 'td[data-stat="losses"]:first',
    'streak': 'td[data-stat="game_streak"]:first'
}

BOXSCORE_SCHEME = {
    'date': 'div[class="scorebox_meta"]',
    'location': 'div[class="scorebox_meta"]',
    'away_name': 'a[itemprop="name"]:first',
    'home_name': 'a[itemprop="name"]:last',
    'winning_name': '',
    'winning_abbr': '',
    'losing_name': '',
    'losing_abbr': '',
    'pace': 'td[data-stat="pace"]:first',
    'away_record': 'div[class="table_wrapper"] h2',
    'away_minutes_played': 'tfoot td[data-stat="mp"]',
    'away_field_goals': 'tfoot td[data-stat="fg"]',
    'away_field_goal_attempts': 'tfoot td[data-stat="fga"]',
    'away_field_goal_percentage': 'tfoot td[data-stat="fg_pct"]',
    'away_two_point_field_goals': 'tfoot td[data-stat="fg2"]',
    'away_two_point_field_goal_attempts': 'tfoot td[data-stat="fg2a"]',
    'away_two_point_field_goal_percentage': 'tfoot td[data-stat="fg2_pct"]',
    'away_three_point_field_goals': 'tfoot td[data-stat="fg3"]',
    'away_three_point_field_goal_attempts': 'tfoot td[data-stat="fg3a"]',
    'away_three_point_field_goal_percentage': 'tfoot td[data-stat="fg3_pct"]',
    'away_free_throws': 'tfoot td[data-stat="ft"]',
    'away_free_throw_attempts': 'tfoot td[data-stat="fta"]',
    'away_free_throw_percentage': 'tfoot td[data-stat="ft_pct"]',
    'away_offensive_rebounds': 'tfoot td[data-stat="orb"]',
    'away_defensive_rebounds': 'tfoot td[data-stat="drb"]',
    'away_total_rebounds': 'tfoot td[data-stat="trb"]',
    'away_assists': 'tfoot td[data-stat="ast"]',
    'away_steals': 'tfoot td[data-stat="stl"]',
    'away_blocks': 'tfoot td[data-stat="blk"]',
    'away_turnovers': 'tfoot td[data-stat="tov"]',
    'away_personal_fouls': 'tfoot td[data-stat="pf"]',
    'away_points': 'tfoot td[data-stat="pts"]',
    'away_true_shooting_percentage': 'tfoot td[data-stat="ts_pct"]',
    'away_effective_field_goal_percentage': 'tfoot td[data-stat="efg_pct"]',
    'away_three_point_attempt_rate': 'tfoot td[data-stat="fg3a_per_fga_pct"]',
    'away_free_throw_attempt_rate': 'tfoot td[data-stat="fta_per_fga_pct"]',
    'away_offensive_rebound_percentage': 'tfoot td[data-stat="orb_pct"]',
    'away_defensive_rebound_percentage': 'tfoot td[data-stat="drb_pct"]',
    'away_total_rebound_percentage': 'tfoot td[data-stat="trb_pct"]',
    'away_assist_percentage': 'tfoot td[data-stat="ast_pct"]',
    'away_steal_percentage': 'tfoot td[data-stat="stl_pct"]',
    'away_block_percentage': 'tfoot td[data-stat="blk_pct"]',
    'away_turnover_percentage': 'tfoot td[data-stat="tov_pct"]',
    'away_offensive_rating': 'tfoot td[data-stat="off_rtg"]',
    'away_defensive_rating': 'tfoot td[data-stat="def_rtg"]',
    'home_record': 'div[class="table_wrapper"] h2',
    'home_minutes_played': 'tfoot td[data-stat="mp"]',
    'home_field_goals': 'tfoot td[data-stat="fg"]',
    'home_field_goal_attempts': 'tfoot td[data-stat="fga"]',
    'home_field_goal_percentage': 'tfoot td[data-stat="fg_pct"]',
    'home_two_point_field_goals': 'tfoot td[data-stat="fg2"]',
    'home_two_point_field_goal_attempts': 'tfoot td[data-stat="fg2a"]',
    'home_two_point_field_goal_percentage': 'tfoot td[data-stat="fg2_pct"]',
    'home_three_point_field_goals': 'tfoot td[data-stat="fg3"]',
    'home_three_point_field_goal_attempts': 'tfoot td[data-stat="fg3a"]',
    'home_three_point_field_goal_percentage': 'tfoot td[data-stat="fg3_pct"]',
    'home_free_throws': 'tfoot td[data-stat="ft"]',
    'home_free_throw_attempts': 'tfoot td[data-stat="fta"]',
    'home_free_throw_percentage': 'tfoot td[data-stat="ft_pct"]',
    'home_offensive_rebounds': 'tfoot td[data-stat="orb"]',
    'home_defensive_rebounds': 'tfoot td[data-stat="drb"]',
    'home_total_rebounds': 'tfoot td[data-stat="trb"]',
    'home_assists': 'tfoot td[data-stat="ast"]',
    'home_steals': 'tfoot td[data-stat="stl"]',
    'home_blocks': 'tfoot td[data-stat="blk"]',
    'home_turnovers': 'tfoot td[data-stat="tov"]',
    'home_personal_fouls': 'tfoot td[data-stat="pf"]',
    'home_points': 'tfoot td[data-stat="pts"]',
    'home_true_shooting_percentage': 'tfoot td[data-stat="ts_pct"]',
    'home_effective_field_goal_percentage': 'tfoot td[data-stat="efg_pct"]',
    'home_three_point_attempt_rate': 'tfoot td[data-stat="fg3a_per_fga_pct"]',
    'home_free_throw_attempt_rate': 'tfoot td[data-stat="fta_per_fga_pct"]',
    'home_offensive_rebound_percentage': 'tfoot td[data-stat="orb_pct"]',
    'home_defensive_rebound_percentage': 'tfoot td[data-stat="drb_pct"]',
    'home_total_rebound_percentage': 'tfoot td[data-stat="trb_pct"]',
    'home_assist_percentage': 'tfoot td[data-stat="ast_pct"]',
    'home_steal_percentage': 'tfoot td[data-stat="stl_pct"]',
    'home_block_percentage': 'tfoot td[data-stat="blk_pct"]',
    'home_turnover_percentage': 'tfoot td[data-stat="tov_pct"]',
    'home_offensive_rating': 'tfoot td[data-stat="off_rtg"]',
    'home_defensive_rating': 'tfoot td[data-stat="def_rtg"]'
}

BOXSCORE_ELEMENT_INDEX = {
    'date': 0,
    'location': 1,
    'home_record': 2,
    'home_minutes_played': 1,
    'home_field_goals': 1,
    'home_field_goal_attempts': 1,
    'home_field_goal_percentage': 1,
    'home_two_point_field_goals': 1,
    'home_two_point_field_goal_attempts': 1,
    'home_two_point_field_goal_percentage': 1,
    'home_three_point_field_goals': 1,
    'home_three_point_field_goal_attempts': 1,
    'home_three_point_field_goal_percentage': 1,
    'home_free_throws': 1,
    'home_free_throw_attempts': 1,
    'home_free_throw_percentage': 1,
    'home_offensive_rebounds': 1,
    'home_defensive_rebounds': 1,
    'home_total_rebounds': 1,
    'home_assists': 1,
    'home_steals': 1,
    'home_blocks': 1,
    'home_turnovers': 1,
    'home_personal_fouls': 1,
    'home_points': 1,
    'home_true_shooting_percentage': 1,
    'home_effective_field_goal_percentage': 1,
    'home_three_point_attempt_rate': 1,
    'home_free_throw_attempt_rate': 1,
    'home_offensive_rebound_percentage': 1,
    'home_defensive_rebound_percentage': 1,
    'home_total_rebound_percentage': 1,
    'home_assist_percentage': 1,
    'home_steal_percentage': 1,
    'home_block_percentage': 1,
    'home_turnover_percentage': 1,
    'home_offensive_rating': 1,
    'home_defensive_rating': 1
}

PLAYER_SCHEME = {
    'season': 'th[data-stat="season"]:first',
    'name': 'h1',
    'team_abbreviation': 'td[data-stat="team_id"]',
    'position': 'td[data-stat="pos"]',
    'height': 'span[itemprop="height"]',
    'weight': 'span[itemprop="weight"]',
    'birth_date': 'td[data-stat=""]',
    'nationality': 'td[data-stat=""]',
    'age': 'nobr',
    'games_played': 'td[data-stat="g"]',
    'games_started': 'td[data-stat="gs"]',
    'minutes_played': 'td[data-stat="mp"]',
    'field_goals': 'td[data-stat="fg"]',
    'field_goal_attempts': 'td[data-stat="fga"]',
    'field_goal_percentage': 'td[data-stat="fg_pct"]',
    'three_pointers': 'td[data-stat="fg3"]',
    'three_point_attempts': 'td[data-stat="fg3a"]',
    'three_point_percentage': 'td[data-stat="fg3_pct"]',
    'two_pointers': 'td[data-stat="fg2"]',
    'two_point_attempts': 'td[data-stat="fg2a"]',
    'two_point_percentage': 'td[data-stat="fg2_pct"]',
    'effective_field_goal_percentage': 'td[data-stat="efg_pct"]',
    'free_throws': 'td[data-stat="ft"]',
    'free_throw_attempts': 'td[data-stat="fta"]',
    'free_throw_percentage': 'td[data-stat="ft_pct"]',
    'offensive_rebounds': 'td[data-stat="orb"]',
    'defensive_rebounds': 'td[data-stat="drb"]',
    'total_rebounds': 'td[data-stat="trb"]',
    'assists': 'td[data-stat="ast"]',
    'steals': 'td[data-stat="stl"]',
    'blocks': 'td[data-stat="blk"]',
    'turnovers': 'td[data-stat="tov"]',
    'personal_fouls': 'td[data-stat="pf"]',
    'points': 'td[data-stat="pts"]',
    'player_efficiency_rating': 'td[data-stat="per"]',
    'true_shooting_percentage': 'td[data-stat="ts_pct"]',
    'three_point_attempt_rate': 'td[data-stat="fg3a_per_fga_pct"]',
    'free_throw_attempt_rate': 'td[data-stat="fta_per_fga_pct"]',
    'offensive_rebound_percentage': 'td[data-stat="orb_pct"]',
    'defensive_rebound_percentage': 'td[data-stat="drb_pct"]',
    'total_rebound_percentage': 'td[data-stat="trb_pct"]',
    'assist_percentage': 'td[data-stat="ast_pct"]',
    'steal_percentage': 'td[data-stat="stl_pct"]',
    'block_percentage': 'td[data-stat="blk_pct"]',
    'turnover_percentage': 'td[data-stat="tov_pct"]',
    'usage_percentage': 'td[data-stat="usg_pct"]',
    'offensive_win_shares': 'td[data-stat="ows"]',
    'defensive_win_shares': 'td[data-stat="dws"]',
    'win_shares': 'td[data-stat="ws"]',
    'win_shares_per_48_minutes': 'td[data-stat="ws_per_48"]',
    'offensive_box_plus_minus': 'td[data-stat="obpm"]',
    'defensive_box_plus_minus': 'td[data-stat="dbpm"]',
    'box_plus_minus': 'td[data-stat="bpm"]',
    'defensive_rating': 'td[data-stat="def_rtg"]',
    'offensive_rating': 'td[data-stat="off_rtg"]',
    'boxscore_box_plus_minus': 'td[data-stat="plus_minus"]',
    'value_over_replacement_player': 'td[data-stat="vorp"]',
    'shooting_distance': 'td[data-stat="avg_dist"]',
    'percentage_shots_two_pointers': 'td[data-stat="fg2a_pct_fga"]',
    'percentage_zero_to_three_footers': 'td[data-stat="pct_fga_00_03"]',
    'percentage_three_to_ten_footers': 'td[data-stat="pct_fga_03_10"]',
    'percentage_ten_to_sixteen_footers': 'td[data-stat="pct_fga_10_16"]',
    'percentage_sixteen_foot_plus_two_pointers':
    'td[data-stat="pct_fga_16_xx"]',
    'percentage_shots_three_pointers': 'td[data-stat="fg3a_pct_fga"]',
    'field_goal_perc_zero_to_three_feet': 'td[data-stat="fg_pct_00_03"]',
    'field_goal_perc_three_to_ten_feet': 'td[data-stat="fg_pct_03_10"]',
    'field_goal_perc_ten_to_sixteen_feet': 'td[data-stat="fg_pct_10_16"]',
    'field_goal_perc_sixteen_foot_plus_two_pointers':
    'td[data-stat="fg_pct_16_xx"]',
    'two_pointers_assisted_percentage': 'td[data-stat="fg2_pct_ast"]',
    'percentage_field_goals_as_dunks': 'td[data-stat="pct_fg2_dunk"]',
    'dunks': 'td[data-stat="fg2_dunk"]',
    'three_pointers_assisted_percentage': 'td[data-stat="fg3_pct_ast"]',
    'percentage_of_three_pointers_from_corner':
    'td[data-stat="pct_fg3a_corner"]',
    'three_point_shot_percentage_from_corner':
    'td[data-stat="fg3_pct_corner"]',
    'half_court_heaves': 'td[data-stat="fg3a_heave"]',
    'half_court_heaves_made': 'td[data-stat="fg3_heave"]',
    'point_guard_percentage': 'td[data-stat="pct_1"]',
    'shooting_guard_percentage': 'td[data-stat="pct_2"]',
    'small_forward_percentage': 'td[data-stat="pct_3"]',
    'power_forward_percentage': 'td[data-stat="pct_4"]',
    'center_percentage': 'td[data-stat="pct_5"]',
    'on_court_plus_minus': 'td[data-stat="plus_minus_on"]',
    'net_plus_minus': 'td[data-stat="plus_minus_net"]',
    'passing_turnovers': 'td[data-stat="tov_bad_pass"]',
    'lost_ball_turnovers': 'td[data-stat="tov_lost_ball"]',
    'other_turnovers': 'td[data-stat="tov_other"]',
    'shooting_fouls': 'td[data-stat="fouls_shooting"]',
    'blocking_fouls': 'td[data-stat="fouls_blocking"]',
    'offensive_fouls': 'td[data-stat="fouls_offensive"]',
    'take_fouls': 'td[data-stat="fouls_take"]',
    'points_generated_by_assists': 'td[data-stat="astd_pts"]',
    'shooting_fouls_drawn': 'td[data-stat="drawn_shooting"]',
    'and_ones': 'td[data-stat="and1s"]',
    'shots_blocked': 'td[data-stat="fga_blkd"]',
    'salary': 'td[data-stat="salary"]'
}

NATIONALITY = {
    'ag': 'Antigua and Barbuda',
    'ar': 'Argentina',
    'au': 'Australia',
    'at': 'Austria',
    'bs': 'Bahamas',
    'be': 'Belgium',
    'ba': 'Bosnia and Herzegovina',
    'br': 'Brazil',
    'bg': 'Bulgaria',
    'cm': 'Cameroon',
    'ca': 'Canada',
    'co': 'Colombia',
    'cv': 'Cape Verde',
    'cn': 'China',
    'hr': 'Croatia',
    'cu': 'Cuba',
    'cz': 'Czech Republic',
    'cd': 'Democratic Replubic of Congo',
    'dk': 'Denmark',
    'dm': 'Dominica',
    'do': 'Dominican Replubic',
    'eg': 'Egypt',
    'ee': 'Estonia',
    'fi': 'Finland',
    'fr': 'France',
    'gf': 'French Guiana',
    'ga': 'Gabon',
    'ge': 'Georgia',
    'de': 'Germany',
    'gh': 'Ghana',
    'gr': 'Greece',
    'gp': 'Guadeloupe',
    'gy': 'Guyana',
    'ht': 'Haiti',
    'hu': 'Hungary',
    'is': 'Iceland',
    'ie': 'Ireland',
    'ir': 'Islamic Replubic of Iran',
    'il': 'Israel',
    'it': 'Italy',
    'jm': 'Jamaica',
    'jp': 'Japan',
    'lv': 'Latvia',
    'lb': 'Lebanon',
    'lt': 'Lithuania',
    'lu': 'Luxembourg',
    'ml': 'Mali',
    'mq': 'Martinique',
    'mx': 'Mexico',
    'me': 'Montenegro',
    'ma': 'Morocco',
    'nl': 'Netherlands',
    'nz': 'New Zealand',
    'ng': 'Nigeria',
    'no': 'Norway',
    'pa': 'Panama',
    'pl': 'Poland',
    'pr': 'Puerto Rico',
    'ke': 'Kenya',
    'kr': 'Republic of Korea',
    'mk': 'Republic of Macedonia',
    'cg': 'Republic of Congo',
    'ro': 'Romania',
    'ru': 'Russian Federation',
    'lc': 'Saint Lucia',
    'vc': 'Saint Vincent and the Grenadines',
    'sd': 'Sudan',
    'sn': 'Senegal',
    'rs': 'Serbia',
    'sk': 'Slovakia',
    'si': 'Slovenia',
    'za': 'South Africa',
    'ss': 'South Sudan',
    'es': 'Spain',
    'se': 'Sweden',
    'ch': 'Switzerland',
    'tw': 'Taiwan',
    'tt': 'Trinidad and Tobago',
    'tn': 'Tunisia',
    'tr': 'Turkey',
    'us': 'United States of America',
    'vi': 'U.S. Virgin Islands',
    'ua': 'Ukraine',
    'gb': 'United Kingdom',
    'tz': 'United Republic of Tanzania',
    'uy': 'Uruguay',
    've': 'Venezuela'
}

SEASON_PAGE_URL = 'http://www.basketball-reference.com/leagues/NBA_%s.html'

SCHEDULE_URL = 'http://www.basketball-reference.com/teams/%s/%s_games.html'

BOXSCORE_URL = 'https://www.basketball-reference.com/boxscores/%s.html'

BOXSCORES_URL = ('https://www.basketball-reference.com/boxscores/'
                 '?month=%s&day=%s&year=%s')

PLAYER_URL = 'https://www.basketball-reference.com/players/%s/%s.html'

ROSTER_URL = 'https://www.basketball-reference.com/teams/%s/%s.html'



import mock
import pytest
from flexmock import flexmock
from os.path import join, dirname
from sportsreference import utils
from sportsreference.ncaaf.conferences import Conference, Conferences


YEAR = 2018


def read_file(filename):
    filepath = join(dirname(__file__), 'ncaaf', filename)
    return open(filepath, 'r').read()


def mock_pyquery(url):
    class MockPQ:
        def __init__(self, html_contents, status_code=200):
            self.url = url
            self.reason = 'Bad URL'  # Used when throwing HTTPErrors
            self.headers = {}  # Used when throwing HTTPErrors
            self.status_code = status_code
            self.html_contents = html_contents
            self.text = html_contents

        def __call__(self, div):
            return read_file()

    if 'BAD' in url:
        return MockPQ('', 404)
    if 'acc' in url:
        html_contents = read_file('%s-acc.html' % YEAR)
        return MockPQ(html_contents)
    if 'sec' in url:
        html_contents = read_file('%s-sec.html' % YEAR)
        return MockPQ(html_contents)
    html_contents = read_file('%s.html' % YEAR)
    return MockPQ(html_contents)


class TestNCAAFConferences:
    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_conferences_integration(self, *args, **kwargs):
        team_conference = {'florida-state': 'acc',
                           'boston-college': 'acc',
                           'clemson': 'acc',
                           'north-carolina-state': 'acc',
                           'syracuse': 'acc',
                           'wake-forest': 'acc',
                           'louisville': 'acc',
                           'virginia-tech': 'acc',
                           'duke': 'acc',
                           'georgia-tech': 'acc',
                           'pittsburgh': 'acc',
                           'virginia': 'acc',
                           'miami-fl': 'acc',
                           'north-carolina': 'acc',
                           'florida': 'sec',
                           'georgia': 'sec',
                           'kentucky': 'sec',
                           'missouri': 'sec',
                           'south-carolina': 'sec',
                           'vanderbilt': 'sec',
                           'tennessee': 'sec',
                           'alabama': 'sec',
                           'arkansas': 'sec',
                           'auburn': 'sec',
                           'louisiana-state': 'sec',
                           'mississippi-state': 'sec',
                           'mississippi': 'sec',
                           'texas-am': 'sec'}
        conferences_result = {'acc': {
                                  'name': 'Atlantic Coast Conference',
                                  'teams': {'florida-state': 'Florida State',
                                            'boston-college': 'Boston College',
                                            'clemson': 'Clemson',
                                            'north-carolina-state':
                                            'North Carolina State',
                                            'syracuse': 'Syracuse',
                                            'wake-forest': 'Wake Forest',
                                            'louisville': 'Louisville',
                                            'virginia-tech': 'Virginia Tech',
                                            'duke': 'Duke',
                                            'georgia-tech': 'Georgia Tech',
                                            'pittsburgh': 'Pitt',
                                            'virginia': 'Virginia',
                                            'miami-fl': 'Miami (FL)',
                                            'north-carolina': 'North Carolina'}
                                    },
                              'sec': {
                                    'name': 'Southeastern Conference',
                                    'teams': {'florida': 'Florida',
                                              'georgia': 'Georgia',
                                              'kentucky': 'Kentucky',
                                              'missouri': 'Missouri',
                                              'south-carolina':
                                              'South Carolina',
                                              'vanderbilt': 'Vanderbilt',
                                              'tennessee': 'Tennessee',
                                              'alabama': 'Alabama',
                                              'arkansas': 'Arkansas',
                                              'auburn': 'Auburn',
                                              'louisiana-state': 'LSU',
                                              'mississippi-state':
                                              'Mississippi State',
                                              'mississippi': 'Ole Miss',
                                              'texas-am': 'Texas A&M'}
                                    }
                              }

        flexmock(utils) \
            .should_receive('_find_year_for_season') \
            .and_return(YEAR)

        conferences = Conferences()

        assert conferences.team_conference == team_conference
        assert conferences.conferences == conferences_result

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_conferences_integration_bad_url(self, *args, **kwargs):
        with pytest.raises(ValueError):
            conferences = Conferences('BAD')

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_conference_integration_bad_url(self, *args, **kwargs):
        with pytest.raises(ValueError):
            conference = Conference('BAD')

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_conference_with_no_names_is_empty(self, *args, **kwargs):
        flexmock(Conference) \
            .should_receive('_get_team_abbreviation') \
            .and_return('')

        conference = Conference('acc')

        assert len(conference._teams) == 0



from flexmock import flexmock
from mock import patch, PropertyMock
from pyquery import PyQuery as pq
from sportsreference import utils
from sportsreference.constants import AWAY, HOME
from sportsreference.nba.boxscore import Boxscore, Boxscores


class MockName:
    def __init__(self, name):
        self._name = name

    def text(self):
        return self._name


class MockField:
    def __init__(self, field):
        self._field = field

    def text(self):
        return self._field


class MockBoxscoreData:
    def __init__(self, fields):
        self._fields = fields

    def __call__(self, field):
        return self

    def items(self):
        return [self._fields]


def mock_pyquery(url):
    class MockPQ:
        def __init__(self, html_contents):
            self.status_code = 404
            self.html_contents = html_contents
            self.text = html_contents

    boxscore = read_file('%s.html' % BOXSCORE)
    return MockPQ(boxscore)


class TestNBABoxscore:
    @patch('requests.get', side_effect=mock_pyquery)
    def setup_method(self, *args, **kwargs):
        flexmock(Boxscore) \
            .should_receive('_parse_game_data') \
            .and_return(None)

        self.boxscore = Boxscore(None)

    def test_away_team_wins(self):
        fake_away_points = PropertyMock(return_value=75)
        fake_home_points = PropertyMock(return_value=70)
        type(self.boxscore)._away_points = fake_away_points
        type(self.boxscore)._home_points = fake_home_points

        assert self.boxscore.winner == AWAY

    def test_home_team_wins(self):
        fake_away_points = PropertyMock(return_value=70)
        fake_home_points = PropertyMock(return_value=75)
        type(self.boxscore)._away_points = fake_away_points
        type(self.boxscore)._home_points = fake_home_points

        assert self.boxscore.winner == HOME

    def test_winning_name_is_home(self):
        expected_name = 'Home Name'

        fake_winner = PropertyMock(return_value=HOME)
        fake_home_name = PropertyMock(return_value=MockName(expected_name))
        type(self.boxscore).winner = fake_winner
        type(self.boxscore)._home_name = fake_home_name

        assert self.boxscore.winning_name == expected_name

    def test_winning_name_is_away(self):
        expected_name = 'Away Name'

        fake_winner = PropertyMock(return_value=AWAY)
        fake_away_name = PropertyMock(return_value=MockName(expected_name))
        type(self.boxscore).winner = fake_winner
        type(self.boxscore)._away_name = fake_away_name

        assert self.boxscore.winning_name == expected_name

    def test_winning_abbr_is_home(self):
        expected_name = 'HOME'

        flexmock(utils) \
            .should_receive('_parse_abbreviation') \
            .and_return(expected_name)

        fake_winner = PropertyMock(return_value=HOME)
        fake_home_abbr = PropertyMock(return_value=MockName(expected_name))
        type(self.boxscore).winner = fake_winner
        type(self.boxscore)._home_abbr = fake_home_abbr

        assert self.boxscore.winning_abbr == expected_name

    def test_winning_abbr_is_away(self):
        expected_name = 'AWAY'

        flexmock(utils) \
            .should_receive('_parse_abbreviation') \
            .and_return(expected_name)

        fake_winner = PropertyMock(return_value=AWAY)
        fake_away_abbr = PropertyMock(return_value=MockName(expected_name))
        type(self.boxscore).winner = fake_winner
        type(self.boxscore)._away_abbr = fake_away_abbr

        assert self.boxscore.winning_abbr == expected_name

    def test_losing_name_is_home(self):
        expected_name = 'Home Name'

        fake_winner = PropertyMock(return_value=AWAY)
        fake_home_name = PropertyMock(return_value=MockName(expected_name))
        type(self.boxscore).winner = fake_winner
        type(self.boxscore)._home_name = fake_home_name

        assert self.boxscore.losing_name == expected_name

    def test_losing_name_is_away(self):
        expected_name = 'Away Name'

        fake_winner = PropertyMock(return_value=HOME)
        fake_away_name = PropertyMock(return_value=MockName(expected_name))
        type(self.boxscore).winner = fake_winner
        type(self.boxscore)._away_name = fake_away_name

        assert self.boxscore.losing_name == expected_name

    def test_losing_abbr_is_home(self):
        expected_name = 'HOME'

        flexmock(utils) \
            .should_receive('_parse_abbreviation') \
            .and_return(expected_name)

        fake_winner = PropertyMock(return_value=AWAY)
        fake_home_abbr = PropertyMock(return_value=MockName(expected_name))
        type(self.boxscore).winner = fake_winner
        type(self.boxscore)._home_abbr = fake_home_abbr

        assert self.boxscore.losing_abbr == expected_name

    def test_losing_abbr_is_away(self):
        expected_name = 'AWAY'

        flexmock(utils) \
            .should_receive('_parse_abbreviation') \
            .and_return(expected_name)

        fake_winner = PropertyMock(return_value=HOME)
        fake_away_abbr = PropertyMock(return_value=MockName(expected_name))
        type(self.boxscore).winner = fake_winner
        type(self.boxscore)._away_abbr = fake_away_abbr

        assert self.boxscore.losing_abbr == expected_name

    def test_invalid_away_record_returns_default_wins(self):
        fake_record = PropertyMock(return_value='Golden State Warriors 1')
        type(self.boxscore)._away_record = fake_record

        assert self.boxscore.away_wins == 0

    def test_invalid_away_record_returns_default_losses(self):
        fake_record = PropertyMock(return_value='Golden State Warriors 1')
        type(self.boxscore)._away_record = fake_record

        assert self.boxscore.away_losses == 0

    def test_invalid_home_record_returns_default_wins(self):
        fake_record = PropertyMock(return_value='Golden State Warriors 1')
        type(self.boxscore)._home_record = fake_record

        assert self.boxscore.home_wins == 0

    def test_invalid_home_record_returns_default_losses(self):
        fake_record = PropertyMock(return_value='Golden State Warriors 1')
        type(self.boxscore)._home_record = fake_record

        assert self.boxscore.home_losses == 0

    def test_invalid_url_returns_none(self):
        result = Boxscore(None)._retrieve_html_page('')

        assert result is None

    def test_no_class_information_returns_dataframe_of_none(self):
        mock_points = PropertyMock(return_value=None)
        type(self.boxscore)._away_points = mock_points
        type(self.boxscore)._home_points = mock_points

        assert self.boxscore.dataframe is None

    def test_nba_game_info(self):
        fields = {
            'date': '7:30 PM, November 9, 2018',
            'location': 'State Farm Arena, Atlanta, Georgia'
        }

        mock_field = """7:30 PM, November 9, 2018
State Farm Arena, Atlanta, Georgia
Logos via Sports Logos.net / About logos
"""

        m = MockBoxscoreData(MockField(mock_field))

        for field, value in fields.items():
            result = self.boxscore._parse_game_date_and_location(field, m)
            assert value == result

    def test_nba_partial_game_info(self):
        fields = {
            'date': '7:30 PM, November 9, 2018',
            'location': None
        }

        mock_field = """7:30 PM, November 9, 2018
Logos via Sports Logos.net / About logos"""

        m = MockBoxscoreData(MockField(mock_field))

        for field, value in fields.items():
            result = self.boxscore._parse_game_date_and_location(field, m)
            assert value == result


class TestNBABoxscores:
    @patch('requests.get', side_effect=mock_pyquery)
    def setup_method(self, *args, **kwargs):
        flexmock(Boxscores) \
            .should_receive('_find_games') \
            .and_return(None)
        self.boxscores = Boxscores(None)

    def test_improper_loser_boxscore_format_skips_game(self):
        flexmock(Boxscores) \
            .should_receive('_get_team_details') \
            .and_return((None, None, None, None, None, None))
        mock_html = pq("""<table class="teams">
<tbody>
    <tr class="loser">
            <td class="right">84</td>
            <td class="right gamelink">
            </td>
    </tr>
    <tr class="winner">
            <td><a href="/teams/IND/2017.html">Indiana</a></td>
            <td class="right">105</td>
            <td class="right">&nbsp;
            </td>
    </tr>
    </tbody>
</table>""")
        games = self.boxscores._extract_game_info([mock_html])

        assert len(games) == 0

    def test_improper_winner_boxscore_format_skips_game(self):
        flexmock(Boxscores) \
            .should_receive('_get_team_details') \
            .and_return((None, None, None, None, None, None))
        mock_html = pq("""<table class="teams">
<tbody>
    <tr class="loser">
            <td><a href="/teams/DET/2017.html">Detroit</a></td>
            <td class="right">84</td>
            <td class="right gamelink">
                    <a href="/boxscores/201702040IND.html">Final</a>
            </td>
    </tr>
    <tr class="winner">
            <td class="right">105</td>
            <td class="right">&nbsp;
            </td>
    </tr>
    </tbody>
</table>""")
        games = self.boxscores._extract_game_info([mock_html])

        assert len(games) == 0

    def test_boxscore_with_no_score_returns_none(self):
        mock_html = pq("""<table class="teams">
<tbody>
    <tr class="loser">
            <td><a href="/teams/DET/2017.html">Detroit</a></td>
            <td class="right gamelink">
                    <a href="/boxscores/201702040IND.html">Final</a>
            </td>
    </tr>
    <tr class="loser">
            <td><a href="/teams/IND/2017.html">Indiana</a></td>
            <td class="right">&nbsp;
            </td>
    </tr>
    </tbody>
</table>""")
        games = self.boxscores._extract_game_info([mock_html])

        assert games == [
            {
                'home_name': 'Indiana',
                'home_abbr': 'IND',
                'away_name': 'Detroit',
                'away_abbr': 'DET',
                'boxscore': '201702040IND',
                'winning_name': None,
                'winning_abbr': None,
                'losing_name': None,
                'losing_abbr': None,
                'home_score': None,
                'away_score': None
            }
        ]



import pandas as pd
import re
from .constants import (ELEMENT_INDEX,
                        PARSING_SCHEME,
                        STANDINGS_URL,
                        TEAM_ELEMENT,
                        TEAM_STATS_URL)
from functools import wraps
from pyquery import PyQuery as pq
from .. import utils
from ..decorators import float_property_decorator, int_property_decorator
from .roster import Roster
from .schedule import Schedule


def mlb_int_property_decorator(func):
    @property
    @wraps(func)
    def wrapper(*args):
        value = func(*args)
        # Equivalent to the calling property's method name
        field = func.__name__
        try:
            record = value.split('-')
        except AttributeError:
            return None
        try:
            return int(record[TEAM_ELEMENT[field]])
        except (TypeError, ValueError, IndexError):
            return None
    return wrapper


class Team(object):
    """
    An object containing all of a team's season information.

    Finds and parses all team stat information and identifiers, such as rank,
    name, and abbreviation, and sets them as properties which can be directly
    read from for easy reference.

    Parameters
    ----------
    team_data : string
        A string containing all of the rows of stats for a given team. If
        multiple tables are being referenced, this will be comprised of
        multiple rows in a single string.
    rank : int
        A team's position in the league based on the number of points they
        obtained during the season.
    year : string (optional)
        The requested year to pull stats from.
    """
    def __init__(self, team_data, rank, year=None):
        self._year = year
        self._rank = rank
        self._abbreviation = None
        self._name = None
        self._league = None
        self._games = None
        self._wins = None
        self._losses = None
        self._win_percentage = None
        self._streak = None
        self._runs = None
        self._runs_against = None
        self._run_difference = None
        self._strength_of_schedule = None
        self._simple_rating_system = None
        self._pythagorean_win_loss = None
        self._luck = None
        self._interleague_record = None
        self._home_record = None
        self._away_record = None
        self._extra_inning_record = None
        self._single_run_record = None
        self._record_vs_right_handed_pitchers = None
        self._record_vs_left_handed_pitchers = None
        self._record_vs_teams_over_500 = None
        self._record_vs_teams_under_500 = None
        self._last_ten_games_record = None
        self._last_twenty_games_record = None
        self._last_thirty_games_record = None
        self._number_players_used = None
        self._average_batter_age = None
        self._plate_appearances = None
        self._at_bats = None
        self._total_runs = None
        self._hits = None
        self._doubles = None
        self._triples = None
        self._home_runs = None
        self._runs_batted_in = None
        self._stolen_bases = None
        self._times_caught_stealing = None
        self._bases_on_balls = None
        self._times_struck_out = None
        self._batting_average = None
        self._on_base_percentage = None
        self._slugging_percentage = None
        self._on_base_plus_slugging_percentage = None
        self._on_base_plus_slugging_percentage_plus = None
        self._total_bases = None
        self._grounded_into_double_plays = None
        self._times_hit_by_pitch = None
        self._sacrifice_hits = None
        self._sacrifice_flies = None
        self._intentional_bases_on_balls = None
        self._runners_left_on_base = None
        self._number_of_pitchers = None
        self._average_pitcher_age = None
        self._runs_allowed_per_game = None
        self._earned_runs_against = None
        self._games_finished = None
        self._complete_games = None
        self._shutouts = None
        self._complete_game_shutouts = None
        self._saves = None
        self._innings_pitched = None
        self._hits_allowed = None
        self._home_runs_against = None
        self._bases_on_walks_given = None
        self._strikeouts = None
        self._hit_pitcher = None
        self._balks = None
        self._wild_pitches = None
        self._batters_faced = None
        self._earned_runs_against_plus = None
        self._fielding_independent_pitching = None
        self._whip = None
        self._hits_per_nine_innings = None
        self._home_runs_per_nine_innings = None
        self._bases_on_walks_given_per_nine_innings = None
        self._strikeouts_per_nine_innings = None
        self._strikeouts_per_base_on_balls = None
        self._opposing_runners_left_on_base = None

        self._parse_team_data(team_data)

    def _parse_name(self, team_data):
        """
        Parses the team's name.

        On the pages being parsed, the team's name doesn't follow the standard
        parsing algorithm that we use for the fields, and requires a special
        one-off algorithm. The name is attached in the 'title' attribute from
        within 'team_ID'. A few simple regex subs captures the team name. The
        '_name' attribute is applied with the captured team name from this
        function.

        Parameters
        ----------
        team_data : string
            A string containing all of the rows of stats for a given team. If
            multiple tables are being referenced, this will be comprised of
            multiple rows in a single string.
        """
        name = team_data('td[data-stat="team_ID"]:first')
        name = re.sub(r'.*title="', '', str(name))
        name = re.sub(r'".*', '', name)
        setattr(self, '_name', name)

    def _parse_team_data(self, team_data):
        """
        Parses a value for every attribute.

        This function looks through every attribute with the exception of
        '_rank' and retrieves the value according to the parsing scheme and
        index of the attribute from the passed HTML data. Once the value is
        retrieved, the attribute's value is updated with the returned result.

        Note that this method is called directly once Team is invoked and does
        not need to be called manually.

        Parameters
        ----------
        team_data : string
            A string containing all of the rows of stats for a given team. If
            multiple tables are being referenced, this will be comprised of
            multiple rows in a single string.
        """
        for field in self.__dict__:
            # The short field truncates the leading '_' in the attribute name.
            short_field = str(field)[1:]
            # The rank attribute is passed directly to the class during
            # instantiation.
            if field == '_rank' or \
               field == '_year':
                continue
            elif field == '_name':
                self._parse_name(team_data)
                continue
            # Default to returning the first element returned unless a
            # subsequent element is desired. For example, total runs and
            # runs per game are two different fields, but they both share
            # the same attribute of 'R' in the HTML tables.
            index = 0
            if short_field in ELEMENT_INDEX.keys():
                index = ELEMENT_INDEX[short_field]
            value = utils._parse_field(PARSING_SCHEME,
                                       team_data,
                                       short_field,
                                       index)
            setattr(self, field, value)

    @property
    def dataframe(self):
        """
        Returns a pandas DataFrame containing all other class properties and
        values. The index for the DataFrame is the string abbreviation of the
        team, such as 'HOU'.
        """
        fields_to_include = {
            'abbreviation': self.abbreviation,
            'at_bats': self.at_bats,
            'average_batter_age': self.average_batter_age,
            'average_pitcher_age': self.average_pitcher_age,
            'away_losses': self.away_losses,
            'away_record': self.away_record,
            'away_wins': self.away_wins,
            'balks': self.balks,
            'bases_on_balls': self.bases_on_balls,
            'bases_on_walks_given': self.bases_on_walks_given,
            'bases_on_walks_given_per_nine_innings':
            self.bases_on_walks_given_per_nine_innings,
            'batters_faced': self.batters_faced,
            'batting_average': self.batting_average,
            'complete_game_shutouts': self.complete_game_shutouts,
            'complete_games': self.complete_games,
            'doubles': self.doubles,
            'earned_runs_against': self.earned_runs_against,
            'earned_runs_against_plus': self.earned_runs_against_plus,
            'extra_inning_losses': self.extra_inning_losses,
            'extra_inning_record': self.extra_inning_record,
            'extra_inning_wins': self.extra_inning_wins,
            'fielding_independent_pitching':
            self.fielding_independent_pitching,
            'games': self.games,
            'games_finished': self.games_finished,
            'grounded_into_double_plays': self.grounded_into_double_plays,
            'hit_pitcher': self.hit_pitcher,
            'hits': self.hits,
            'hits_allowed': self.hits_allowed,
            'hits_per_nine_innings': self.hits_per_nine_innings,
            'home_losses': self.home_losses,
            'home_record': self.home_record,
            'home_runs': self.home_runs,
            'home_runs_against': self.home_runs_against,
            'home_runs_per_nine_innings': self.home_runs_per_nine_innings,
            'home_wins': self.home_wins,
            'innings_pitched': self.innings_pitched,
            'intentional_bases_on_balls': self.intentional_bases_on_balls,
            'interleague_record': self.interleague_record,
            'last_ten_games_record': self.last_ten_games_record,
            'last_thirty_games_record': self.last_thirty_games_record,
            'last_twenty_games_record': self.last_twenty_games_record,
            'league': self.league,
            'losses': self.losses,
            'losses_last_ten_games': self.losses_last_ten_games,
            'losses_last_thirty_games': self.losses_last_thirty_games,
            'losses_last_twenty_games': self.losses_last_twenty_games,
            'losses_vs_left_handed_pitchers':
            self.losses_vs_left_handed_pitchers,
            'losses_vs_right_handed_pitchers':
            self.losses_vs_right_handed_pitchers,
            'losses_vs_teams_over_500': self.losses_vs_teams_over_500,
            'losses_vs_teams_under_500': self.losses_vs_teams_under_500,
            'luck': self.luck,
            'name': self.name,
            'number_of_pitchers': self.number_of_pitchers,
            'number_players_used': self.number_players_used,
            'on_base_percentage': self.on_base_percentage,
            'on_base_plus_slugging_percentage':
            self.on_base_plus_slugging_percentage,
            'on_base_plus_slugging_percentage_plus':
            self.on_base_plus_slugging_percentage_plus,
            'opposing_runners_left_on_base':
            self.opposing_runners_left_on_base,
            'plate_appearances': self.plate_appearances,
            'pythagorean_win_loss': self.pythagorean_win_loss,
            'rank': self.rank,
            'record_vs_left_handed_pitchers':
            self.record_vs_left_handed_pitchers,
            'record_vs_right_handed_pitchers':
            self.record_vs_right_handed_pitchers,
            'record_vs_teams_over_500': self.record_vs_teams_over_500,
            'record_vs_teams_under_500': self.record_vs_teams_under_500,
            'run_difference': self.run_difference,
            'runners_left_on_base': self.runners_left_on_base,
            'runs': self.runs,
            'runs_against': self.runs_against,
            'runs_allowed_per_game': self.runs_allowed_per_game,
            'runs_batted_in': self.runs_batted_in,
            'sacrifice_flies': self.sacrifice_flies,
            'sacrifice_hits': self.sacrifice_hits,
            'saves': self.saves,
            'shutouts': self.shutouts,
            'simple_rating_system': self.simple_rating_system,
            'single_run_losses': self.single_run_losses,
            'single_run_record': self.single_run_record,
            'single_run_wins': self.single_run_wins,
            'slugging_percentage': self.slugging_percentage,
            'stolen_bases': self.stolen_bases,
            'streak': self.streak,
            'strength_of_schedule': self.strength_of_schedule,
            'strikeouts': self.strikeouts,
            'strikeouts_per_base_on_balls': self.strikeouts_per_base_on_balls,
            'strikeouts_per_nine_innings': self.strikeouts_per_nine_innings,
            'times_caught_stealing': self.times_caught_stealing,
            'times_hit_by_pitch': self.times_hit_by_pitch,
            'times_struck_out': self.times_struck_out,
            'total_bases': self.total_bases,
            'total_runs': self.total_runs,
            'triples': self.triples,
            'whip': self.whip,
            'wild_pitches': self.wild_pitches,
            'win_percentage': self.win_percentage,
            'wins': self.wins,
            'wins_last_ten_games': self.wins_last_ten_games,
            'wins_last_thirty_games': self.wins_last_thirty_games,
            'wins_last_twenty_games': self.wins_last_twenty_games,
            'wins_vs_left_handed_pitchers': self.wins_vs_left_handed_pitchers,
            'wins_vs_right_handed_pitchers':
            self.wins_vs_right_handed_pitchers,
            'wins_vs_teams_over_500': self.wins_vs_teams_over_500,
            'wins_vs_teams_under_500': self.wins_vs_teams_under_500
        }
        return pd.DataFrame([fields_to_include], index=[self._abbreviation])

    @int_property_decorator
    def rank(self):
        """
        Returns an ``int`` of the team's rank based on their win percentage.
        """
        return self._rank

    @property
    def abbreviation(self):
        """
        Returns a ``string`` of the team's abbreviation, such as 'HOU' for the
        Houston Astros.
        """
        return self._abbreviation

    @property
    def schedule(self):
        """
        Returns an instance of the Schedule class containing the team's
        complete schedule for the season.
        """
        return Schedule(self._abbreviation, self._year)

    @property
    def roster(self):
        """
        Returns an instance of the Roster class containing all players for the
        team during the season with all career stats.
        """
        return Roster(self._abbreviation, self._year)

    @property
    def name(self):
        """
        Returns a ``string`` of the team's full name, such as 'Houston Astros'.
        """
        return self._name

    @property
    def league(self):
        """
        Returns a ``string`` of the two letter abbreviation of the league, such
        as 'AL' for the American League.
        """
        return self._league

    @int_property_decorator
    def games(self):
        """
        Returns an ``int`` of the number of games the team has played during
        the season.
        """
        return self._games

    @int_property_decorator
    def wins(self):
        """
        Returns an ``int`` of the total number of games the team won during the
        season.
        """
        return self._wins

    @int_property_decorator
    def losses(self):
        """
        Returns an ``int`` of the total number of games the team lost during
        the season.
        """
        return self._losses

    @float_property_decorator
    def win_percentage(self):
        """
        Returns a ``float`` of the number of wins divided by the number of
        games played during the season. Percentage ranges from 0-1.
        """
        return self._win_percentage

    @property
    def streak(self):
        """
        Returns a ``string`` of the team's current winning or losing streak,
        such as 'W 3' for a team on a 3-game winning streak.
        """
        return self._streak

    @float_property_decorator
    def runs(self):
        """
        Returns a ``float`` of the average number of runs scored per game by
        the team.
        """
        return self._runs

    @float_property_decorator
    def runs_against(self):
        """
        Returns a ``float`` of the average number of runs scored per game by
        the opponent.
        """
        return self._runs_against

    @float_property_decorator
    def run_difference(self):
        """
        Returns a ``float`` of the difference between the number of runs scored
        and the number of runs given up per game. Positive numbers indicate
        the team scores more per game than they are scored on.
        """
        return self._run_difference

    @float_property_decorator
    def strength_of_schedule(self):
        """
        Returns a ``float`` denoting a team's strength of schedule, based on
        runs scores and conceded. Higher values result in more challenging
        schedules while 0.0 is an average schedule.
        """
        return self._strength_of_schedule

    @float_property_decorator
    def simple_rating_system(self):
        """
        Returns a ``float`` of the average number of runs per game a team
        scores compared to average.
        """
        return self._simple_rating_system

    @property
    def pythagorean_win_loss(self):
        """
        Returns a ``string`` of the team's expected win-loss record based on
        the runs scored and allowed. Record is in the format 'W-L'.
        """
        return self._pythagorean_win_loss

    @int_property_decorator
    def luck(self):
        """
        Returns an ``int`` of the difference between the current wins and
        losses compared to the pythagorean wins and losses.
        """
        return self._luck

    @property
    def interleague_record(self):
        """
        Returns a ``string`` of the team's interleague record. Record is in the
        format 'W-L'.
        """
        return self._interleague_record

    @property
    def home_record(self):
        """
        Returns a ``string`` of the team's home record. Record is in the format
        'W-L'.
        """
        return self._home_record

    @mlb_int_property_decorator
    def home_wins(self):
        """
        Returns an ``int`` of the number of wins at home during the season.
        """
        return self._home_record

    @mlb_int_property_decorator
    def home_losses(self):
        """
        Returns an ``int`` of the number of losses at home during the season.
        """
        return self._home_record

    @property
    def away_record(self):
        """
        Returns a ``string`` of the team's away record. Record is in the format
        'W-L'.
        """
        return self._away_record

    @mlb_int_property_decorator
    def away_wins(self):
        """
        Returns an ``int`` of the number of away wins during the season.
        """
        return self._away_record

    @mlb_int_property_decorator
    def away_losses(self):
        """
        Returns an ``int`` of the number of away losses during the season.
        """
        return self._away_record

    @property
    def extra_inning_record(self):
        """
        Returns a ``string`` of the team's record when the game has gone to
        extra innings. Record is in the format 'W-L'.
        """
        return self._extra_inning_record

    @mlb_int_property_decorator
    def extra_inning_wins(self):
        """
        Returns an ``int`` of the number of wins the team has when the game has
        gone to extra innings.
        """
        return self._extra_inning_record

    @mlb_int_property_decorator
    def extra_inning_losses(self):
        """
        Returns an ``int`` of the number of losses the team has when the game
        has gone to extra innings.
        """
        return self._extra_inning_record

    @property
    def single_run_record(self):
        """
        Returns a ``string`` of the team's record when only one run is scored.
        Record is in the format 'W-L'.
        """
        return self._single_run_record

    @mlb_int_property_decorator
    def single_run_wins(self):
        """
        Returns an ``int`` of the number of wins the team has when only one run
        is scored.
        """
        return self._single_run_record

    @mlb_int_property_decorator
    def single_run_losses(self):
        """
        Returns an ``int`` of the number of losses the team has when only one
        run is scored.
        """
        return self._single_run_record

    @property
    def record_vs_right_handed_pitchers(self):
        """
        Returns a ``string`` of the team's record against right-handed
        pitchers.
        Record is in the format 'W-L'.
        """
        return self._record_vs_right_handed_pitchers

    @mlb_int_property_decorator
    def wins_vs_right_handed_pitchers(self):
        """
        Returns an ``int`` of the number of wins against right-handed pitchers.
        """
        return self._record_vs_right_handed_pitchers

    @mlb_int_property_decorator
    def losses_vs_right_handed_pitchers(self):
        """
        Returns an ``int`` of the number of losses against right-handed
        pitchers.
        """
        return self._record_vs_right_handed_pitchers

    @property
    def record_vs_left_handed_pitchers(self):
        """
        Returns a ``string`` of the team's record against left-handed pitchers.
        Record is in the format 'W-L'.
        """
        return self._record_vs_left_handed_pitchers

    @mlb_int_property_decorator
    def wins_vs_left_handed_pitchers(self):
        """
        Returns an ``int`` of number of wins against left-handed pitchers.
        """
        return self._record_vs_left_handed_pitchers

    @mlb_int_property_decorator
    def losses_vs_left_handed_pitchers(self):
        """
        Returns an ``int`` of number of losses against left-handed pitchers.
        """
        return self._record_vs_left_handed_pitchers

    @property
    def record_vs_teams_over_500(self):
        """
        Returns a ``string`` of the team's record against teams with a win
        percentage over 500. Record is in the format 'W-L'.
        """
        return self._record_vs_teams_over_500

    @mlb_int_property_decorator
    def wins_vs_teams_over_500(self):
        """
        Returns an ``int`` of the number of wins against teams over 500.
        """
        return self._record_vs_teams_over_500

    @mlb_int_property_decorator
    def losses_vs_teams_over_500(self):
        """
        Returns an ``int`` of the number of losses against teams over 500.
        """
        return self._record_vs_teams_over_500

    @property
    def record_vs_teams_under_500(self):
        """
        Returns a ``string`` of the team's record against teams with a win
        percentage under 500. Record is in the format 'W-L'.
        """
        return self._record_vs_teams_under_500

    @mlb_int_property_decorator
    def wins_vs_teams_under_500(self):
        """
        Returns an ``int`` of the number of wins against teams under 500.
        """
        return self._record_vs_teams_under_500

    @mlb_int_property_decorator
    def losses_vs_teams_under_500(self):
        """
        Returns an ``int`` of the number of losses against teams under 500.
        """
        return self._record_vs_teams_under_500

    @property
    def last_ten_games_record(self):
        """
        Returns a ``string`` of the team's record over the last ten games.
        Record is in the format 'W-L'.
        """
        return self._last_ten_games_record

    @mlb_int_property_decorator
    def wins_last_ten_games(self):
        """
        Returns an ``int`` of the number of wins in the last 10 games.
        """
        return self._last_ten_games_record

    @mlb_int_property_decorator
    def losses_last_ten_games(self):
        """
        Returns an ``int`` of the number of losses in the last 10 games.
        """
        return self._last_ten_games_record

    @property
    def last_twenty_games_record(self):
        """
        Returns a ``string`` of the team's record over the last twenty games.
        Record is in the format 'W-L'.
        """
        return self._last_twenty_games_record

    @mlb_int_property_decorator
    def wins_last_twenty_games(self):
        """
        Returns an ``int`` of the number of wins in the last 20 games.
        """
        return self._last_twenty_games_record

    @mlb_int_property_decorator
    def losses_last_twenty_games(self):
        """
        Returns an ``int`` of the number of losses in the last 20 games.
        """
        return self._last_twenty_games_record

    @property
    def last_thirty_games_record(self):
        """
        Returns a ``string`` of the team's record over the last thirty games.
        Record is in the format 'W-L'.
        """
        return self._last_thirty_games_record

    @mlb_int_property_decorator
    def wins_last_thirty_games(self):
        """
        Returns an ``int`` of the number of wins in the last 30 games.
        """
        return self._last_thirty_games_record

    @mlb_int_property_decorator
    def losses_last_thirty_games(self):
        """
        Returns an ``int`` of the number of losses in the last 30 games.
        """
        return self._last_thirty_games_record

    @int_property_decorator
    def number_players_used(self):
        """
        Returns an ``int`` of the number of different players used during the
        season.
        """
        return self._number_players_used

    @float_property_decorator
    def average_batter_age(self):
        """
        Returns a ``float`` of the average batter age weighted by their number
        of at bats plus the number of games participated in.
        """
        return self._average_batter_age

    @int_property_decorator
    def plate_appearances(self):
        """
        Returns an ``int`` of the total number of plate appearances for the
        team.
        """
        return self._plate_appearances

    @int_property_decorator
    def at_bats(self):
        """
        Returns an ``int`` of the total number of at bats for the team.
        """
        return self._at_bats

    @int_property_decorator
    def total_runs(self):
        """
        Returns an ``int`` of the total number of runs scored during the
        season.
        """
        return self._total_runs

    @int_property_decorator
    def hits(self):
        """
        Returns an ``int`` of the total number of hits during the season.
        """
        return self._hits

    @int_property_decorator
    def doubles(self):
        """
        Returns an ``int`` of the total number of doubles hit by the team.
        """
        return self._doubles

    @int_property_decorator
    def triples(self):
        """
        Returns an ``int`` of the total number of tripes hit by the team.
        """
        return self._triples

    @int_property_decorator
    def home_runs(self):
        """
        Returns an ``int`` of the total number of home runs hit by the team.
        """
        return self._home_runs

    @int_property_decorator
    def runs_batted_in(self):
        """
        Returns an ``int`` of the total number of runs batted in by the team.
        """
        return self._runs_batted_in

    @int_property_decorator
    def stolen_bases(self):
        """
        Returns an ``int`` of the total number of bases stolen by the team.
        """
        return self._stolen_bases

    @int_property_decorator
    def times_caught_stealing(self):
        """
        Returns an ``int`` of the number of times a player was caught stealing.
        """
        return self._times_caught_stealing

    @int_property_decorator
    def bases_on_balls(self):
        """
        Returns an ``int`` of the number of bases on walks.
        """
        return self._bases_on_balls

    @int_property_decorator
    def times_struck_out(self):
        """
        Returns an ``int`` of the total number of times the team struck out.
        """
        return self._times_struck_out

    @float_property_decorator
    def batting_average(self):
        """
        Returns a ``float`` of the batting average for the team. Percentage
        ranges from 0-1.
        """
        return self._batting_average

    @float_property_decorator
    def on_base_percentage(self):
        """
        Returns a ``float`` of the percentage of at bats that result in a
        player taking a base. Percentage ranges from 0-1.
        """
        return self._on_base_percentage

    @float_property_decorator
    def slugging_percentage(self):
        """
        Returns a ``float`` of the ratio of total bases gained per at bat.
        """
        return self._slugging_percentage

    @float_property_decorator
    def on_base_plus_slugging_percentage(self):
        """
        Returns a ``float`` of the sum of the on base percentage plus the
        slugging percentage.
        """
        return self._on_base_plus_slugging_percentage

    @int_property_decorator
    def on_base_plus_slugging_percentage_plus(self):
        """
        Returns an ``int`` of the on base percentage plus the slugging
        percentage, adjusted to the team's home ballpark.
        """
        return self._on_base_plus_slugging_percentage_plus

    @int_property_decorator
    def total_bases(self):
        """
        Returns an ``int`` of the total number of bases a team has gained
        during the season.
        """
        return self._total_bases

    @int_property_decorator
    def grounded_into_double_plays(self):
        """
        Returns an ``int`` of the total number double plays grounded into by
        the team.
        """
        return self._grounded_into_double_plays

    @int_property_decorator
    def times_hit_by_pitch(self):
        """
        Returns an ``int`` of the total number of times a batter was hit by an
        opponent's pitch.
        """
        return self._times_hit_by_pitch

    @int_property_decorator
    def sacrifice_hits(self):
        """
        Returns an ``int`` of the total number of sacrifice hits the team made
        during the season.
        """
        return self._sacrifice_hits

    @int_property_decorator
    def sacrifice_flies(self):
        """
        Returns an ``int`` of the total number of sacrifice flies the team made
        during the season.
        """
        return self._sacrifice_flies

    @int_property_decorator
    def intentional_bases_on_balls(self):
        """
        Returns an ``int`` of the total number of times a player took a base
        from an intentional walk.
        """
        return self._intentional_bases_on_balls

    @int_property_decorator
    def runners_left_on_base(self):
        """
        Returns an ``int`` of the total number of runners left on base at the
        end of an inning.
        """
        return self._runners_left_on_base

    @int_property_decorator
    def number_of_pitchers(self):
        """
        Returns an ``int`` of the total number of pitchers used during a
        season.
        """
        return self._number_of_pitchers

    @float_property_decorator
    def average_pitcher_age(self):
        """
        Returns a ``float`` of the average pitcher age weighted by the number
        of games started, followed by the number of games played and saves.
        """
        return self._average_pitcher_age

    @float_property_decorator
    def runs_allowed_per_game(self):
        """
        Returns a ``float`` of the average number of runs a team has allowed
        per game.
        """
        return self._runs_allowed_per_game

    @float_property_decorator
    def earned_runs_against(self):
        """
        Returns a ``float`` of the average number of earned runs against for a
        team.
        """
        return self._earned_runs_against

    @int_property_decorator
    def games_finished(self):
        """
        Returns an ``int`` of the number of games finished which is equivalent
        to the number of games played minus the number of complete games during
        the season.
        """
        return self._games_finished

    @int_property_decorator
    def complete_games(self):
        """
        Returns an ``int`` of the total number of complete games a team has
        accumulated during the season.
        """
        return self._complete_games

    @int_property_decorator
    def shutouts(self):
        """
        Returns an ``int`` of the total number of shutouts a team has
        accumulated during the season.
        """
        return self._shutouts

    @int_property_decorator
    def complete_game_shutouts(self):
        """
        Returns an ``int`` of the total number of complete games where the
        opponent scored zero runs.
        """
        return self._complete_game_shutouts

    @int_property_decorator
    def saves(self):
        """
        Returns an ``int`` of the total number of saves a team has accumulated
        during the season.
        """
        return self._saves

    @float_property_decorator
    def innings_pitched(self):
        """
        Returns a ``float`` of the total number of innings pitched by a team
        during the season.
        """
        return self._innings_pitched

    @int_property_decorator
    def hits_allowed(self):
        """
        Returns an ``int`` of the total number of hits allowed during the
        season.
        """
        return self._hits_allowed

    @int_property_decorator
    def home_runs_against(self):
        """
        Returns an ``int`` of the total number of home runs given up during the
        season.
        """
        return self._home_runs_against

    @int_property_decorator
    def bases_on_walks_given(self):
        """
        Returns an ``int`` of the total number of bases from walks given up by
        a team during the season.
        """
        return self._bases_on_walks_given

    @int_property_decorator
    def strikeouts(self):
        """
        Returns an ``int`` of the total number of times a team has struck out
        an opponent.
        """
        return self._strikeouts

    @int_property_decorator
    def hit_pitcher(self):
        """
        Returns an ``int`` of the total number of times a pitcher has hit an
        opposing batter.
        """
        return self._hit_pitcher

    @int_property_decorator
    def balks(self):
        """
        Returns an ``int`` of the total number of times a pitcher has balked.
        """
        return self._balks

    @int_property_decorator
    def wild_pitches(self):
        """
        Returns an ``int`` of the total number of wild pitches thrown by a team
        during a season.
        """
        return self._wild_pitches

    @int_property_decorator
    def batters_faced(self):
        """
        Returns an ``int`` of the total number of batters all pitchers have
        faced during a season.
        """
        return self._batters_faced

    @int_property_decorator
    def earned_runs_against_plus(self):
        """
        Returns an ``int`` of the team's average earned runs against, adjusted
        for the home ballpark.
        """
        return self._earned_runs_against_plus

    @float_property_decorator
    def fielding_independent_pitching(self):
        """
        Returns a ``float`` of the team's effectiveness at preventing home
        runs, walks, batters being hit by pitches, and strikeouts.
        """
        return self._fielding_independent_pitching

    @float_property_decorator
    def whip(self):
        """
        Returns a ``float`` of the average number of walks plus hits by the
        opponent per inning.
        """
        return self._whip

    @float_property_decorator
    def hits_per_nine_innings(self):
        """
        Returns a ``float`` of the average number of hits per nine innings by
        the opponent.
        """
        return self._hits_per_nine_innings

    @float_property_decorator
    def home_runs_per_nine_innings(self):
        """
        Returns a ``float`` of the average number of home runs per nine innings
        by the opponent.
        """
        return self._home_runs_per_nine_innings

    @float_property_decorator
    def bases_on_walks_given_per_nine_innings(self):
        """
        Returns a ``float`` of the average number of walks conceded per nine
        innings.
        """
        return self._bases_on_walks_given_per_nine_innings

    @float_property_decorator
    def strikeouts_per_nine_innings(self):
        """
        Returns a ``float`` of the average number of strikeouts a team throws
        per nine innings.
        """
        return self._strikeouts_per_nine_innings

    @float_property_decorator
    def strikeouts_per_base_on_balls(self):
        """
        Returns a ``float`` of the average number of strikeouts per walk thrown
        by a team.
        """
        return self._strikeouts_per_base_on_balls

    @int_property_decorator
    def opposing_runners_left_on_base(self):
        """
        Returns an ``int`` of the total number of opponents a team has left on
        bases at the end of an inning.
        """
        return self._opposing_runners_left_on_base


class Teams:
    """
    A list of all MLB teams and their stats in a given year.

    Finds and retrieves a list of all MLB teams from www.baseball-reference.com
    and creates a Team instance for every team that participated in the league
    in a given year. The Team class comprises a list of all major stats and a
    few identifiers for the requested season.

    Parameters
    ----------
    year : string (optional)
        The requested year to pull stats from.
    """
    def __init__(self, year=None):
        self._teams = []

        self._retrieve_all_teams(year)

    def __getitem__(self, abbreviation):
        """
        Return a specified team.

        Returns a team's instance in the Teams class as specified by the team's
        abbreviation.

        Parameters
        ----------
        abbreviation : string
            An MLB team's three letter abbreviation (ie. 'HOU' for Houston
            Astros).

        Returns
        -------
        Team instance
            If the requested team can be found, its Team instance is returned.

        Raises
        ------
        ValueError
            If the requested team is not present within the Teams list.
        """
        for team in self._teams:
            if team.abbreviation.upper() == abbreviation.upper():
                return team
        raise ValueError('Team abbreviation %s not found' % abbreviation)

    def __call__(self, abbreviation):
        """
        Return a specified team.

        Returns a team's instance in the Teams class as specified by the team's
        abbreviation. This method is a wrapper for __getitem__.

        Parameters
        ----------
        abbreviation : string
            An MLB team's three letter abbreviation (ie. 'HOU' for Houson
            Astros).

        Returns
        -------
        Team instance
            If the requested team can be found, its Team instance is returned.
        """
        return self.__getitem__(abbreviation)

    def __repr__(self):
        """Returns a ``list`` of all MLB teams for the given season."""
        return self._teams

    def __iter__(self):
        """Returns an iterator of all of the MLB teams for a given season."""
        return iter(self.__repr__())

    def __len__(self):
        """Returns the number of MLB teams for a given season."""
        return len(self.__repr__())

    def _add_stats_data(self, teams_list, team_data_dict):
        """
        Add a team's stats row to a dictionary.

        Pass table contents and a stats dictionary of all teams to accumulate
        all stats for each team in a single variable.

        Parameters
        ----------
        teams_list : generator
            A generator of all row items in a given table.
        team_data_dict : {str: {'data': str, 'rank': int}} dictionary
            A dictionary where every key is the team's abbreviation and every
            value is another dictionary with a 'data' key which contains the
            string version of the row data for the matched team, and a 'rank'
            key which is the rank of the team.

        Returns
        -------
        dictionary
            An updated version of the team_data_dict with the passed table row
            information included.
        """
        # Teams are listed in terms of rank with the first team being #1
        rank = 1
        for team_data in teams_list:
            # Skip the league average row
            if 'class="league_average_table"' in str(team_data):
                continue
            abbr = utils._parse_field(PARSING_SCHEME,
                                      team_data,
                                      'abbreviation')
            try:
                team_data_dict[abbr]['data'] += team_data
            except KeyError:
                team_data_dict[abbr] = {'data': team_data, 'rank': rank}
            rank += 1
        return team_data_dict

    def _retrieve_all_teams(self, year):
        """
        Find and create Team instances for all teams in the given season.

        For a given season, parses the specified MLB stats table and finds all
        requested stats. Each team then has a Team instance created which
        includes all requested stats and a few identifiers, such as the team's
        name and abbreviation. All of the individual Team instances are added
        to a list.

        Note that this method is called directly once Teams is invoked and does
        not need to be called manually.

        Parameters
        ----------
        year : string
            The requested year to pull stats from.
        """
        team_data_dict = {}

        if not year:
            year = utils._find_year_for_season('mlb')
        doc = pq(STANDINGS_URL % year)
        div_prefix = 'div#all_expanded_standings_overall'
        standings = utils._get_stats_table(doc, div_prefix)
        doc = pq(TEAM_STATS_URL % year)
        div_prefix = 'div#all_teams_standard_%s'
        batting_stats = utils._get_stats_table(doc, div_prefix % 'batting')
        pitching_stats = utils._get_stats_table(doc, div_prefix % 'pitching')
        for stats_list in [standings, batting_stats, pitching_stats]:
            team_data_dict = self._add_stats_data(stats_list, team_data_dict)

        for team_data in team_data_dict.values():
            team = Team(team_data['data'], team_data['rank'], year)
            self._teams.append(team)

    @property
    def dataframes(self):
        """
        Returns a pandas DataFrame where each row is a representation of the
        Team class. Rows are indexed by the team abbreviation.
        """
        frames = []
        for team in self.__iter__():
            frames.append(team.dataframe)
        return pd.concat(frames)



from flexmock import flexmock
from mock import PropertyMock
from sportsreference.constants import (AWAY,
                                       HOME,
                                       LOSS,
                                       WIN)
from sportsreference.nba.schedule import Game, Schedule


class TestNBASchedule:
    def setup_method(self, *args, **kwargs):
        flexmock(Game) \
            .should_receive('_parse_game_data') \
            .and_return(None)

        self.game = Game(None)

    def test_away_game_returns_away_location(self):
        fake_location = PropertyMock(return_value='@')
        type(self.game)._location = fake_location

        assert self.game.location == AWAY

    def test_home_game_returns_home_location(self):
        fake_location = PropertyMock(return_value='')
        type(self.game)._location = fake_location

        assert self.game.location == HOME

    def test_winning_result_returns_win(self):
        fake_result = PropertyMock(return_value='W')
        type(self.game)._result = fake_result

        assert self.game.result == WIN

    def test_losing_result_returns_loss(self):
        fake_result = PropertyMock(return_value='L')
        type(self.game)._result = fake_result

        assert self.game.result == LOSS

    def test_empty_game_class_returns_dataframe_of_none(self):
        assert self.game._points_allowed is None
        assert self.game._points_scored is None
        assert self.game.dataframe is None

    def test_no_dataframes_returns_none(self):
        flexmock(Schedule) \
            .should_receive('_pull_schedule') \
            .and_return(None)
        schedule = Schedule('DET')

        fake_game = flexmock(dataframe=None)
        fake_games = PropertyMock(return_value=fake_game)
        type(schedule).__iter__ = fake_games

        assert schedule.dataframe is None

    def test_no_dataframes_extended_returns_none(self):
        flexmock(Schedule) \
            .should_receive('_pull_schedule') \
            .and_return(None)
        schedule = Schedule('DET')

        fake_game = flexmock(dataframe_extended=None)
        fake_games = PropertyMock(return_value=fake_game)
        type(schedule).__iter__ = fake_games

        assert schedule.dataframe_extended is None



import mock
import os
import pandas as pd
from datetime import datetime
from flexmock import flexmock
from sportsreference import utils
from sportsreference.constants import AWAY
from sportsreference.nhl.constants import BOXSCORE_URL, BOXSCORES_URL
from sportsreference.nhl.boxscore import Boxscore, Boxscores


MONTH = 10
YEAR = 2017

BOXSCORE = '201806070VEG'


def read_file(filename):
    filepath = os.path.join(os.path.dirname(__file__), 'nhl', filename)
    return open('%s' % filepath, 'r').read()


def mock_pyquery(url):
    class MockPQ:
        def __init__(self, html_contents):
            self.status_code = 200
            self.html_contents = html_contents
            self.text = html_contents

    if url == BOXSCORES_URL % (2, 4, YEAR):
        return MockPQ(read_file('boxscores-2-4-2017.html'))
    if url == BOXSCORES_URL % (2, 5, YEAR):
        return MockPQ(read_file('boxscores-2-5-2017.html'))
    boxscore = read_file('%s.html' % BOXSCORE)
    return MockPQ(boxscore)


class MockDateTime:
    def __init__(self, year, month):
        self.year = year
        self.month = month


class TestNHLBoxscore:
    @mock.patch('requests.get', side_effect=mock_pyquery)
    def setup_method(self, *args, **kwargs):
        self.results = {
            'date': 'June 7, 2018',
            'time': '8:00 PM',
            'arena': 'T-Mobile Arena',
            'attendance': 18529,
            'duration': '2:45',
            'winner': AWAY,
            'winning_name': 'Washington Capitals',
            'winning_abbr': 'WSH',
            'losing_name': 'Vegas Golden Knights',
            'losing_abbr': 'VEG',
            'away_goals': 4,
            'away_assists': 7,
            'away_points': 11,
            'away_penalties_in_minutes': 8,
            'away_even_strength_goals': 3,
            'away_power_play_goals': 1,
            'away_short_handed_goals': 0,
            'away_game_winning_goals': 1,
            'away_even_strength_assists': 5,
            'away_power_play_assists': 2,
            'away_short_handed_assists': 0,
            'away_shots_on_goal': 33,
            'away_shooting_percentage': 12.1,
            'away_saves': 28,
            'away_save_percentage': .903,
            'away_shutout': 0,
            'home_goals': 3,
            'home_assists': 6,
            'home_points': 9,
            'home_penalties_in_minutes': 12,
            'home_even_strength_goals': 2,
            'home_power_play_goals': 1,
            'home_short_handed_goals': 0,
            'home_game_winning_goals': 0,
            'home_even_strength_assists': 4,
            'home_power_play_assists': 2,
            'home_short_handed_assists': 0,
            'home_shots_on_goal': 31,
            'home_shooting_percentage': 9.7,
            'home_saves': 29,
            'home_save_percentage': .879,
            'home_shutout': 0
        }

        flexmock(utils) \
            .should_receive('_todays_date') \
            .and_return(MockDateTime(YEAR, MONTH))

        self.boxscore = Boxscore(BOXSCORE)

    def test_nhl_boxscore_returns_requested_boxscore(self):
        for attribute, value in self.results.items():
            assert getattr(self.boxscore, attribute) == value

    def test_invalid_url_yields_empty_class(self):
        flexmock(Boxscore) \
            .should_receive('_retrieve_html_page') \
            .and_return(None)

        boxscore = Boxscore(BOXSCORE)

        for key, value in boxscore.__dict__.items():
            if key == '_uri':
                continue
            assert value is None

    def test_nhl_boxscore_dataframe_returns_dataframe_of_all_values(self):
        df = pd.DataFrame([self.results], index=[BOXSCORE])

        # Pandas doesn't natively allow comparisons of DataFrames.
        # Concatenating the two DataFrames (the one generated during the test
        # and the expected one above) and dropping duplicate rows leaves only
        # the rows that are unique between the two frames. This allows a quick
        # check of the DataFrame to see if it is empty - if so, all rows are
        # duplicates, and they are equal.
        frames = [df, self.boxscore.dataframe]
        df1 = pd.concat(frames).drop_duplicates(keep=False)

        assert df1.empty

    def test_nhl_boxscore_player(self):
        boxscore = Boxscore(BOXSCORE)

        assert len(boxscore.home_players) == 19
        assert len(boxscore.away_players) == 19

        for player in boxscore.home_players:
            assert not player.dataframe.empty
        for player in boxscore.away_players:
            assert not player.dataframe.empty


class TestNHLBoxscores:
    def setup_method(self):
        self.expected = {
            '2-4-2017': [
                {'boxscore': '201702040BOS',
                 'away_name': 'Toronto Maple Leafs',
                 'away_abbr': 'TOR',
                 'away_score': 6,
                 'home_name': 'Boston Bruins',
                 'home_abbr': 'BOS',
                 'home_score': 5,
                 'winning_name': 'Toronto Maple Leafs',
                 'winning_abbr': 'TOR',
                 'losing_name': 'Boston Bruins',
                 'losing_abbr': 'BOS'},
                {'boxscore': '201702040BUF',
                 'away_name': 'Ottawa Senators',
                 'away_abbr': 'OTT',
                 'away_score': 0,
                 'home_name': 'Buffalo Sabres',
                 'home_abbr': 'BUF',
                 'home_score': 4,
                 'winning_name': 'Buffalo Sabres',
                 'winning_abbr': 'BUF',
                 'losing_name': 'Ottawa Senators',
                 'losing_abbr': 'OTT'},
                {'boxscore': '201702040CBJ',
                 'away_name': 'New Jersey Devils',
                 'away_abbr': 'NJD',
                 'away_score': 5,
                 'home_name': 'Columbus Blue Jackets',
                 'home_abbr': 'CBJ',
                 'home_score': 1,
                 'winning_name': 'New Jersey Devils',
                 'winning_abbr': 'NJD',
                 'losing_name': 'Columbus Blue Jackets',
                 'losing_abbr': 'CBJ'},
                {'boxscore': '201702040COL',
                 'away_name': 'Winnipeg Jets',
                 'away_abbr': 'WPG',
                 'away_score': 2,
                 'home_name': 'Colorado Avalanche',
                 'home_abbr': 'COL',
                 'home_score': 5,
                 'winning_name': 'Colorado Avalanche',
                 'winning_abbr': 'COL',
                 'losing_name': 'Winnipeg Jets',
                 'losing_abbr': 'WPG'},
                {'boxscore': '201702040DAL',
                 'away_name': 'Chicago Blackhawks',
                 'away_abbr': 'CHI',
                 'away_score': 5,
                 'home_name': 'Dallas Stars',
                 'home_abbr': 'DAL',
                 'home_score': 3,
                 'winning_name': 'Chicago Blackhawks',
                 'winning_abbr': 'CHI',
                 'losing_name': 'Dallas Stars',
                 'losing_abbr': 'DAL'},
                {'boxscore': '201702040MTL',
                 'away_name': 'Washington Capitals',
                 'away_abbr': 'WSH',
                 'away_score': 3,
                 'home_name': 'Montreal Canadiens',
                 'home_abbr': 'MTL',
                 'home_score': 2,
                 'winning_name': 'Washington Capitals',
                 'winning_abbr': 'WSH',
                 'losing_name': 'Montreal Canadiens',
                 'losing_abbr': 'MTL'},
                {'boxscore': '201702040NSH',
                 'away_name': 'Detroit Red Wings',
                 'away_abbr': 'DET',
                 'away_score': 1,
                 'home_name': 'Nashville Predators',
                 'home_abbr': 'NSH',
                 'home_score': 0,
                 'winning_name': 'Detroit Red Wings',
                 'winning_abbr': 'DET',
                 'losing_name': 'Nashville Predators',
                 'losing_abbr': 'NSH'},
                {'boxscore': '201702040NYI',
                 'away_name': 'Carolina Hurricanes',
                 'away_abbr': 'CAR',
                 'away_score': 5,
                 'home_name': 'New York Islanders',
                 'home_abbr': 'NYI',
                 'home_score': 4,
                 'winning_name': 'Carolina Hurricanes',
                 'winning_abbr': 'CAR',
                 'losing_name': 'New York Islanders',
                 'losing_abbr': 'NYI'},
                {'boxscore': '201702040PHI',
                 'away_name': 'Los Angeles Kings',
                 'away_abbr': 'LAK',
                 'away_score': 1,
                 'home_name': 'Philadelphia Flyers',
                 'home_abbr': 'PHI',
                 'home_score': 0,
                 'winning_name': 'Los Angeles Kings',
                 'winning_abbr': 'LAK',
                 'losing_name': 'Philadelphia Flyers',
                 'losing_abbr': 'PHI'},
                {'boxscore': '201702040SJS',
                 'away_name': 'Arizona Coyotes',
                 'away_abbr': 'ARI',
                 'away_score': 3,
                 'home_name': 'San Jose Sharks',
                 'home_abbr': 'SJS',
                 'home_score': 2,
                 'winning_name': 'Arizona Coyotes',
                 'winning_abbr': 'ARI',
                 'losing_name': 'San Jose Sharks',
                 'losing_abbr': 'SJS'},
                {'boxscore': '201702040STL',
                 'away_name': 'Pittsburgh Penguins',
                 'away_abbr': 'PIT',
                 'away_score': 4,
                 'home_name': 'St. Louis Blues',
                 'home_abbr': 'STL',
                 'home_score': 1,
                 'winning_name': 'Pittsburgh Penguins',
                 'winning_abbr': 'PIT',
                 'losing_name': 'St. Louis Blues',
                 'losing_abbr': 'STL'},
                {'boxscore': '201702040TBL',
                 'away_name': 'Anaheim Ducks',
                 'away_abbr': 'ANA',
                 'away_score': 2,
                 'home_name': 'Tampa Bay Lightning',
                 'home_abbr': 'TBL',
                 'home_score': 3,
                 'winning_name': 'Tampa Bay Lightning',
                 'winning_abbr': 'TBL',
                 'losing_name': 'Anaheim Ducks',
                 'losing_abbr': 'ANA'},
                {'boxscore': '201702040VAN',
                 'away_name': 'Minnesota Wild',
                 'away_abbr': 'MIN',
                 'away_score': 6,
                 'home_name': 'Vancouver Canucks',
                 'home_abbr': 'VAN',
                 'home_score': 3,
                 'winning_name': 'Minnesota Wild',
                 'winning_abbr': 'MIN',
                 'losing_name': 'Vancouver Canucks',
                 'losing_abbr': 'VAN'}
            ]
        }

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_boxscores_search(self, *args, **kwargs):
        result = Boxscores(datetime(2017, 2, 4)).games

        assert result == self.expected

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_boxscores_search_invalid_end(self, *args, **kwargs):
        result = Boxscores(datetime(2017, 2, 4), datetime(2017, 2, 3)).games

        assert result == self.expected

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_boxscores_search_multiple_days(self, *args, **kwargs):
        expected = {
            '2-4-2017': [
                {'boxscore': '201702040BOS',
                 'away_name': 'Toronto Maple Leafs',
                 'away_abbr': 'TOR',
                 'away_score': 6,
                 'home_name': 'Boston Bruins',
                 'home_abbr': 'BOS',
                 'home_score': 5,
                 'winning_name': 'Toronto Maple Leafs',
                 'winning_abbr': 'TOR',
                 'losing_name': 'Boston Bruins',
                 'losing_abbr': 'BOS'},
                {'boxscore': '201702040BUF',
                 'away_name': 'Ottawa Senators',
                 'away_abbr': 'OTT',
                 'away_score': 0,
                 'home_name': 'Buffalo Sabres',
                 'home_abbr': 'BUF',
                 'home_score': 4,
                 'winning_name': 'Buffalo Sabres',
                 'winning_abbr': 'BUF',
                 'losing_name': 'Ottawa Senators',
                 'losing_abbr': 'OTT'},
                {'boxscore': '201702040CBJ',
                 'away_name': 'New Jersey Devils',
                 'away_abbr': 'NJD',
                 'away_score': 5,
                 'home_name': 'Columbus Blue Jackets',
                 'home_abbr': 'CBJ',
                 'home_score': 1,
                 'winning_name': 'New Jersey Devils',
                 'winning_abbr': 'NJD',
                 'losing_name': 'Columbus Blue Jackets',
                 'losing_abbr': 'CBJ'},
                {'boxscore': '201702040COL',
                 'away_name': 'Winnipeg Jets',
                 'away_abbr': 'WPG',
                 'away_score': 2,
                 'home_name': 'Colorado Avalanche',
                 'home_abbr': 'COL',
                 'home_score': 5,
                 'winning_name': 'Colorado Avalanche',
                 'winning_abbr': 'COL',
                 'losing_name': 'Winnipeg Jets',
                 'losing_abbr': 'WPG'},
                {'boxscore': '201702040DAL',
                 'away_name': 'Chicago Blackhawks',
                 'away_abbr': 'CHI',
                 'away_score': 5,
                 'home_name': 'Dallas Stars',
                 'home_abbr': 'DAL',
                 'home_score': 3,
                 'winning_name': 'Chicago Blackhawks',
                 'winning_abbr': 'CHI',
                 'losing_name': 'Dallas Stars',
                 'losing_abbr': 'DAL'},
                {'boxscore': '201702040MTL',
                 'away_name': 'Washington Capitals',
                 'away_abbr': 'WSH',
                 'away_score': 3,
                 'home_name': 'Montreal Canadiens',
                 'home_abbr': 'MTL',
                 'home_score': 2,
                 'winning_name': 'Washington Capitals',
                 'winning_abbr': 'WSH',
                 'losing_name': 'Montreal Canadiens',
                 'losing_abbr': 'MTL'},
                {'boxscore': '201702040NSH',
                 'away_name': 'Detroit Red Wings',
                 'away_abbr': 'DET',
                 'away_score': 1,
                 'home_name': 'Nashville Predators',
                 'home_abbr': 'NSH',
                 'home_score': 0,
                 'winning_name': 'Detroit Red Wings',
                 'winning_abbr': 'DET',
                 'losing_name': 'Nashville Predators',
                 'losing_abbr': 'NSH'},
                {'boxscore': '201702040NYI',
                 'away_name': 'Carolina Hurricanes',
                 'away_abbr': 'CAR',
                 'away_score': 5,
                 'home_name': 'New York Islanders',
                 'home_abbr': 'NYI',
                 'home_score': 4,
                 'winning_name': 'Carolina Hurricanes',
                 'winning_abbr': 'CAR',
                 'losing_name': 'New York Islanders',
                 'losing_abbr': 'NYI'},
                {'boxscore': '201702040PHI',
                 'away_name': 'Los Angeles Kings',
                 'away_abbr': 'LAK',
                 'away_score': 1,
                 'home_name': 'Philadelphia Flyers',
                 'home_abbr': 'PHI',
                 'home_score': 0,
                 'winning_name': 'Los Angeles Kings',
                 'winning_abbr': 'LAK',
                 'losing_name': 'Philadelphia Flyers',
                 'losing_abbr': 'PHI'},
                {'boxscore': '201702040SJS',
                 'away_name': 'Arizona Coyotes',
                 'away_abbr': 'ARI',
                 'away_score': 3,
                 'home_name': 'San Jose Sharks',
                 'home_abbr': 'SJS',
                 'home_score': 2,
                 'winning_name': 'Arizona Coyotes',
                 'winning_abbr': 'ARI',
                 'losing_name': 'San Jose Sharks',
                 'losing_abbr': 'SJS'},
                {'boxscore': '201702040STL',
                 'away_name': 'Pittsburgh Penguins',
                 'away_abbr': 'PIT',
                 'away_score': 4,
                 'home_name': 'St. Louis Blues',
                 'home_abbr': 'STL',
                 'home_score': 1,
                 'winning_name': 'Pittsburgh Penguins',
                 'winning_abbr': 'PIT',
                 'losing_name': 'St. Louis Blues',
                 'losing_abbr': 'STL'},
                {'boxscore': '201702040TBL',
                 'away_name': 'Anaheim Ducks',
                 'away_abbr': 'ANA',
                 'away_score': 2,
                 'home_name': 'Tampa Bay Lightning',
                 'home_abbr': 'TBL',
                 'home_score': 3,
                 'winning_name': 'Tampa Bay Lightning',
                 'winning_abbr': 'TBL',
                 'losing_name': 'Anaheim Ducks',
                 'losing_abbr': 'ANA'},
                {'boxscore': '201702040VAN',
                 'away_name': 'Minnesota Wild',
                 'away_abbr': 'MIN',
                 'away_score': 6,
                 'home_name': 'Vancouver Canucks',
                 'home_abbr': 'VAN',
                 'home_score': 3,
                 'winning_name': 'Minnesota Wild',
                 'winning_abbr': 'MIN',
                 'losing_name': 'Vancouver Canucks',
                 'losing_abbr': 'VAN'}
            ],
            '2-5-2017': [
                {'boxscore': '201702050MTL',
                 'away_name': 'Edmonton Oilers',
                 'away_abbr': 'EDM',
                 'away_score': 1,
                 'home_name': 'Montreal Canadiens',
                 'home_abbr': 'MTL',
                 'home_score': 0,
                 'winning_name': 'Edmonton Oilers',
                 'winning_abbr': 'EDM',
                 'losing_name': 'Montreal Canadiens',
                 'losing_abbr': 'MTL'},
                {'boxscore': '201702050NYR',
                 'away_name': 'Calgary Flames',
                 'away_abbr': 'CGY',
                 'away_score': 3,
                 'home_name': 'New York Rangers',
                 'home_abbr': 'NYR',
                 'home_score': 4,
                 'winning_name': 'New York Rangers',
                 'winning_abbr': 'NYR',
                 'losing_name': 'Calgary Flames',
                 'losing_abbr': 'CGY'},
                {'boxscore': '201702050WSH',
                 'away_name': 'Los Angeles Kings',
                 'away_abbr': 'LAK',
                 'away_score': 0,
                 'home_name': 'Washington Capitals',
                 'home_abbr': 'WSH',
                 'home_score': 5,
                 'winning_name': 'Washington Capitals',
                 'winning_abbr': 'WSH',
                 'losing_name': 'Los Angeles Kings',
                 'losing_abbr': 'LAK'}
            ]
        }
        result = Boxscores(datetime(2017, 2, 4), datetime(2017, 2, 5)).games

        assert result == expected



from flexmock import flexmock
from mock import PropertyMock
from pyquery import PyQuery as pq
from sportsreference import utils
from sportsreference.constants import (AWAY,
                                       CONFERENCE_TOURNAMENT,
                                       HOME,
                                       LOSS,
                                       NEUTRAL,
                                       NON_DI,
                                       REGULAR_SEASON,
                                       WIN)
from sportsreference.ncaab.constants import (CBI_TOURNAMENT,
                                             CIT_TOURNAMENT,
                                             NCAA_TOURNAMENT,
                                             NIT_TOURNAMENT)
from sportsreference.ncaab.schedule import Game, Schedule


class TestNCAABSchedule:
    def setup_method(self, *args, **kwargs):
        flexmock(Game) \
            .should_receive('_parse_game_data') \
            .and_return(None)

        self.game = Game(None)

    def test_away_game_returns_away_location(self):
        fake_location = PropertyMock(return_value='@')
        type(self.game)._location = fake_location

        assert self.game.location == AWAY

    def test_home_game_returns_home_location(self):
        fake_location = PropertyMock(return_value='')
        type(self.game)._location = fake_location

        assert self.game.location == HOME

    def test_neutral_game_returns_neutral_location(self):
        fake_location = PropertyMock(return_value='N')
        type(self.game)._location = fake_location

        assert self.game.location == NEUTRAL

    def test_winning_result_returns_win(self):
        fake_result = PropertyMock(return_value='W')
        type(self.game)._result = fake_result

        assert self.game.result == WIN

    def test_losing_result_returns_loss(self):
        fake_result = PropertyMock(return_value='L')
        type(self.game)._result = fake_result

        assert self.game.result == LOSS

    def test_regular_season_game_type(self):
        fake_type = PropertyMock(return_value='REG')
        type(self.game)._type = fake_type

        assert self.game.type == REGULAR_SEASON

    def test_conference_tournament_game_type(self):
        fake_type = PropertyMock(return_value='CTOURN')
        type(self.game)._type = fake_type

        assert self.game.type == CONFERENCE_TOURNAMENT

    def test_ncaa_tournament_game_type(self):
        fake_type = PropertyMock(return_value='NCAA')
        type(self.game)._type = fake_type

        assert self.game.type == NCAA_TOURNAMENT

    def test_nit_tournament_game_type(self):
        fake_type = PropertyMock(return_value='NIT')
        type(self.game)._type = fake_type

        assert self.game.type == NIT_TOURNAMENT

    def test_cbi_tournament_game_type(self):
        fake_type = PropertyMock(return_value='CBI')
        type(self.game)._type = fake_type

        assert self.game.type == CBI_TOURNAMENT

    def test_cit_tournament_game_type(self):
        fake_type = PropertyMock(return_value='CIT')
        type(self.game)._type = fake_type

        assert self.game.type == CIT_TOURNAMENT

    def test_team_with_rank_returns_rank(self):
        fake_name = PropertyMock(return_value='Purdue (3)')
        type(self.game)._opponent_name = fake_name

        assert self.game.opponent_rank == 3

    def test_team_with_no_rank_returns_none(self):
        fake_name = PropertyMock(return_value='Kansas State')
        type(self.game)._opponent_name = fake_name

        assert self.game.opponent_rank is None

    def test_overtimes_returns_number_of_overtimes(self):
        fake_overtime = PropertyMock(return_value='OT')
        type(self.game)._overtimes = fake_overtime

        assert self.game.overtimes == 1

    def test_multiple_overtimes_returns_number_of_overtimes(self):
        fake_overtime = PropertyMock(return_value='2OT')
        type(self.game)._overtimes = fake_overtime

        assert self.game.overtimes == 2

    def test_no_overtime_returns_zero(self):
        fake_overtime = PropertyMock(return_value='')
        type(self.game)._overtimes = fake_overtime

        assert self.game.overtimes == 0

    def test_bad_overtime_defaults_to_zero(self):
        fake_overtime = PropertyMock(return_value='BAD')
        type(self.game)._overtimes = fake_overtime

        assert self.game.overtimes == 0

    def test_empty_schedule_class_returns_dataframe_of_none(self):
        fake_points = PropertyMock(return_value=None)
        type(self.game)._home_points = fake_points
        type(self.game)._away_points = fake_points

        assert self.game._home_points is None
        assert self.game._away_points is None
        assert self.game.dataframe is None

    def test_empty_schedule_class_returns_dataframe_extended_of_none(self):
        fake_points = PropertyMock(return_value=None)
        type(self.game)._home_points = fake_points
        type(self.game)._away_points = fake_points

        assert self.game._home_points is None
        assert self.game._away_points is None
        assert self.game.dataframe_extended is None

    def test_no_dataframes_returns_none(self):
        flexmock(Schedule) \
            .should_receive('_pull_schedule') \
            .and_return(None)
        schedule = Schedule('PURDUE')

        fake_game = flexmock(dataframe=None)
        fake_games = PropertyMock(return_value=fake_game)
        type(schedule).__iter__ = fake_games

        assert schedule.dataframe is None

    def test_no_dataframes_extended_returns_none(self):
        flexmock(Schedule) \
            .should_receive('_pull_schedule') \
            .and_return(None)
        schedule = Schedule('PURDUE')

        fake_game = flexmock(dataframe_extended=None)
        fake_games = PropertyMock(return_value=fake_game)
        type(schedule).__iter__ = fake_games

        assert schedule.dataframe_extended is None


class TestNCAABScheduleNames:
    def test_non_major_school_returns_name_for_abbreviation(self):
        text = ('<td class="left " data-stat="opp_name">'
                'City College of New York</td>')
        game_data = pq(text)

        game = Game(game_data)

        assert game.opponent_abbr == 'City College of New York'

    def test_non_major_school_returns_non_dI_for_conference(self):
        game_data = pq('<td class="left " data-stat="conf_abbr"></td>')

        game = Game(game_data)

        assert game.opponent_conference == NON_DI
